# Copyright 2021 - 2022, Martijn Braam and the OpenAtem contributors
# SPDX-License-Identifier: GPL-3.0-only
import gettext
import logging
import re
from datetime import datetime, timedelta
from logging import Logger

from gtk_switcher.decorators import field
from gtk_switcher.layout import LayoutView
from gtk_switcher.panel_color import ColorPanel
from gtk_switcher.panel_mediaplayer import MediaPlayerPanel
from gtk_switcher.panel_supersource import SupersourcePanel
from pyatem.command import CutCommand, AutoCommand, FadeToBlackCommand, TransitionSettingsCommand, WipeSettingsCommand, \
    TransitionPositionCommand, TransitionPreviewCommand, ColorGeneratorCommand, MixSettingsCommand, DipSettingsCommand, \
    DveSettingsCommand, AudioMasterPropertiesCommand, FairlightMasterPropertiesCommand, DkeyRateCommand, \
    DkeyAutoCommand, DkeyTieCommand, \
    DkeyOnairCommand, ProgramInputCommand, PreviewInputCommand, KeyOnAirCommand, KeyFillCommand, \
    FadeToBlackConfigCommand, RecorderStatusCommand, AuxSourceCommand, StreamingServiceSetCommand, \
    RecordingSettingsSetCommand, StreamingStatusSetCommand, MediaplayerSelectCommand, StreamingAudioBitrateCommand, \
    MacroRecordCommand, MacroActionCommand
from pyatem.field import TransitionSettingsField, InputPropertiesField, TopologyField, RecordingSettingsField
import gtk_switcher.stream_data

import gi

gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, GLib, GObject, Gio, Gdk

gi.require_version('Handy', '1')
from gi.repository import Handy


class SwitcherPage:
    def __init__(self, builder):
        self.main_blocks = builder.get_object('main_blocks')
        self.me = []
        self.layout = {}

        self.log_sw = logging.getLogger('SwitcherPage')

        self.mix_rate = builder.get_object('mix_rate')
        self.dip_rate = builder.get_object('dip_rate')
        self.wipe_rate = builder.get_object('wipe_rate')
        self.dve_rate = builder.get_object('dve_rate')
        self.ftb_rate = builder.get_object('ftb_rate')
        self.dip_source = builder.get_object('dip_source')
        self.wipe_symmetry_adj = builder.get_object('wipe_symmetry_adj')
        self.wipe_x_adj = builder.get_object('wipe_x_adj')
        self.wipe_y_adj = builder.get_object('wipe_y_adj')
        self.wipe_width_adj = builder.get_object('wipe_width_adj')
        self.wipe_softness_adj = builder.get_object('wipe_softness_adj')
        self.wipe_fill = builder.get_object('wipe_fill')
        self.wipe_reverse = builder.get_object('wipe_reverse')
        self.wipe_flipflop = builder.get_object('wipe_flipflop')

        self.dve_reverse = builder.get_object('dve_reverse')
        self.dve_flipflop = builder.get_object('dve_flipflop')

        self.ftb_afv = builder.get_object('ftb_afv')

        self.keyer_stack = builder.get_object('keyer_stack')

        self.usks = {}
        self.dsks = {}
        self.has_models = []
        self.menu = None

        self.upstream_keyers = builder.get_object('upstream_keyers')
        self.downstream_keyers = builder.get_object('downstream_keyers')
        self.macro_flow = builder.get_object('macro_flow')

        self.usk1_dve_fill = builder.get_object('usk1_dve_fill')
        self.usk1_mask_en = builder.get_object('usk1_mask_en')
        self.usk1_mask_top = builder.get_object('usk1_mask_top')
        self.usk1_mask_bottom = builder.get_object('usk1_mask_bottom')
        self.usk1_mask_left = builder.get_object('usk1_mask_left')
        self.usk1_mask_right = builder.get_object('usk1_mask_right')

        self.expander_stream_recorder = builder.get_object('expander_stream_recorder')
        self.expander_livestream = builder.get_object('expander_livestream')
        self.expander_encoder = builder.get_object('expander_encoder')
        self.stream_recorder_filename = builder.get_object('stream_recorder_filename')
        self.stream_recorder_disk2 = builder.get_object('stream_recorder_disk2')
        self.stream_recorder_disk1 = builder.get_object('stream_recorder_disk1')
        self.stream_recorder_disk1_label = builder.get_object('stream_recorder_disk1_label')
        self.stream_recorder_disk2_label = builder.get_object('stream_recorder_disk2_label')
        self.stream_recorder_start = builder.get_object('stream_recorder_start')
        self.stream_recorder_stop = builder.get_object('stream_recorder_stop')
        self.stream_recorder_switch = builder.get_object('stream_recorder_switch')
        self.stream_recorder_clock = builder.get_object('stream_recorder_clock')
        self.stream_recorder_status = builder.get_object('stream_recorder_status')
        self.stream_recorder_trigger_cameras = builder.get_object('stream_recorder_trigger_cameras')
        self.stream_recorder_disk = [None, None]
        self.stream_recorder_active = False
        self.stream_recorder_start_time = None

        self.audio_rate_min = builder.get_object('audio_rate_min')
        self.audio_rate_max = builder.get_object('audio_rate_max')
        self.video_rate_min = builder.get_object('video_rate_min')
        self.video_rate_max = builder.get_object('video_rate_max')

        self.stream_presets = builder.get_object('stream_presets')
        self.stream_live_clock = builder.get_object('stream_live_clock')
        self.stream_live_status = builder.get_object('stream_live_status')
        self.stream_live_platform = builder.get_object('stream_live_platform')
        self.stream_live_server = builder.get_object('stream_live_server')
        self.stream_live_key = builder.get_object('stream_live_key')
        self.stream_live_start = builder.get_object('stream_live_start')
        self.stream_live_stop = builder.get_object('stream_live_stop')
        self.live_stats = builder.get_object('live_stats')
        self.stream_live_active = False
        self.stream_live_start_time = None

        self.macro_name = builder.get_object('macro_name')

        self.flap = builder.get_object('flap')
        self.flaptoggle = builder.get_object('flaptoggle')

        self.flap.bind_property("reveal-flap", self.flaptoggle, "active", GObject.BindingFlags.BIDIRECTIONAL)

        self.switcher_mediaplayers = builder.get_object('switcher_mediaplayers')

        action_streampreset = Gio.SimpleAction.new("streampreset", GLib.VariantType.new("a{sv}"))
        action_streampreset.connect("activate", self.load_livestream_preset)
        self.application.add_action(action_streampreset)

        self.create_livestream_presets()

        self.disks = {}
        self.aux = {}

        self.grid_aux = builder.get_object('grid_aux')

        self.wipe_style = [
            builder.get_object('wipestyle_h'),
            builder.get_object('wipestyle_v'),
            builder.get_object('wipestyle_mh'),
            builder.get_object('wipestyle_mv'),
            builder.get_object('wipestyle_cross'),
            builder.get_object('wipestyle_box'),
            builder.get_object('wipestyle_diamond'),
            builder.get_object('wipestyle_iris'),
            builder.get_object('wipestyle_box_tl'),
            builder.get_object('wipestyle_box_tr'),
            builder.get_object('wipestyle_box_br'),
            builder.get_object('wipestyle_box_bl'),
            builder.get_object('wipestyle_box_top'),
            builder.get_object('wipestyle_box_right'),
            builder.get_object('wipestyle_box_bottom'),
            builder.get_object('wipestyle_box_left'),
            builder.get_object('wipestyle_diag1'),
            builder.get_object('wipestyle_diag2'),
        ]

        for style, button in enumerate(self.wipe_style):
            button.pattern = style
            button.connect('pressed', self.on_wipe_pattern_clicked)

        directions = ['tl', 't', 'tr', 'l', 'r', 'bl', 'b', 'br']
        self.dve_style_push = []
        self.dve_style_squeeze = []
        for stylecode, direction in enumerate(directions):
            btn = builder.get_object(f'dt_push_{direction}')
            btn.style = 'push'
            btn.direction = direction
            btn.styleindex = stylecode + 24
            btn.connect('pressed', self.on_dve_transition_style_clicked)
            self.dve_style_push.append(btn)

            btn = builder.get_object(f'dt_squeeze_{direction}')
            btn.style = 'squeeze'
            btn.direction = direction
            btn.styleindex = stylecode + 16
            btn.connect('pressed', self.on_dve_transition_style_clicked)
            self.dve_style_squeeze.append(btn)

        self.palette_top = builder.get_object('palette_top')

        self.model_me1_fill = builder.get_object('model_me1_fill')
        self.model_key = builder.get_object('model_key')
        self.model_aux = builder.get_object('model_aux')
        self.model_disks = builder.get_object('model_disks')
        self.model_supersource_box = Gtk.ListStore()
        self.model_supersource_box.set_column_types([str, str])
        self.model_supersource_art = Gtk.ListStore()
        self.model_supersource_art.set_column_types([str, str])
        self.model_changing = False
        self.slider_held = False

    def add_mixeffect(self):
        from gtk_switcher.mixeffect import MixEffectBlock
        index = len(self.me)
        me = MixEffectBlock(index)
        self.me.append(me)
        me.set_dsk(False)
        self.main_blocks.add(me)

        me.connect('program-changed', self.on_me_program_changed)
        me.connect('preview-changed', self.on_me_preview_changed)
        me.connect('rate-focus', self.on_rate_focus)
        me.connect('rate-unfocus', self.on_rate_unfocus)
        me.connect('ftb-clicked', self.on_ftb_clicked)
        me.connect('ftb-rate', self.on_ftb_rate_changed)
        me.connect('tbar-position-changed', self.on_tbar_position_changed)
        me.connect('auto-rate-changed', self.on_auto_rate_changed)
        me.connect('auto-clicked', self.on_auto_clicked)
        me.connect('cut-clicked', self.on_cut_clicked)
        me.connect('preview-transition-clicked', self.on_prev_trans_clicked)
        me.connect('style-changed', self.on_style_clicked)
        me.connect('onair-clicked', self.on_onair_clicked)
        me.connect('next-clicked', self.on_next_clicked)
        me.connect('dsk-tie', self.on_dsk_tie_clicked)
        me.connect('dsk-onair', self.on_dsk_onair_clicked)
        me.connect('dsk-auto', self.on_dsk_auto_clicked)
        me.connect('dsk-rate', self.on_dsk_rate_activate)

        layout = LayoutView(index, self.connection)
        self.layout[index] = layout
        layout_exp = Gtk.Expander(label=_("Layout editor M/E {}").format(index + 1))
        layout_exp.set_margin_start(12)
        layout_exp.set_margin_end(12)
        layout_exp.add(layout)
        layout_exp.show_all()
        self.main_blocks.add(layout_exp)

    def on_cut_clicked(self, widget, index):
        cmd = CutCommand(index=index)
        self.connection.mixer.send_commands([cmd])

    def on_cut_shortcut(self, *args):
        if self.disable_shortcuts:
            return

        cmd = CutCommand(index=0)
        self.connection.mixer.send_commands([cmd])

    def on_auto_clicked(self, widget, index, *args):
        if self.disable_shortcuts and len(args) == 2:
            return

        cmd = AutoCommand(index=index)
        self.connection.mixer.send_commands([cmd])

    def on_auto_shortcut(self, *args):
        if self.disable_shortcuts:
            return

        cmd = AutoCommand(index=0)
        self.connection.mixer.send_commands([cmd])

    def on_ftb_clicked(self, widget, index):
        cmd = FadeToBlackCommand(index=index)
        self.connection.mixer.send_commands([cmd])

    def on_ftb_rate_changed(self, widget, index, frames):
        cmd = FadeToBlackConfigCommand(index=index, frames=frames)
        self.connection.mixer.send_commands([cmd])

    def on_style_clicked(self, widget, index, style):
        s = None
        if style == 'mix':
            s = TransitionSettingsField.STYLE_MIX
        elif style == 'dip':
            s = TransitionSettingsField.STYLE_DIP
        elif style == 'wipe':
            s = TransitionSettingsField.STYLE_WIPE
        elif style == 'sting':
            s = TransitionSettingsField.STYLE_STING
        elif style == 'dve':
            s = TransitionSettingsField.STYLE_DVE
        cmd = TransitionSettingsCommand(index=index, style=s)
        self.connection.mixer.send_commands([cmd])

    def on_wipe_pattern_clicked(self, widget):
        if self.model_changing:
            return
        cmd = WipeSettingsCommand(index=0, pattern=widget.pattern)
        self.connection.mixer.send_commands([cmd])

    def on_dve_transition_style_clicked(self, widget):
        if self.model_changing:
            return

        cmd = DveSettingsCommand(index=0, style=widget.styleindex)
        self.connection.mixer.send_commands([cmd])

    def on_dve_reverse_clicked(self, widget):
        if self.model_changing:
            return
        state = widget.get_style_context().has_class('active')
        cmd = DveSettingsCommand(index=0, reverse=not state)
        self.connection.mixer.send_commands([cmd])

    def on_dve_flipflop_clicked(self, widget):
        if self.model_changing:
            return
        state = widget.get_style_context().has_class('active')
        cmd = DveSettingsCommand(index=0, flipflop=not state)
        self.connection.mixer.send_commands([cmd])

    def on_tbar_position_changed(self, widget, index, position):
        cmd = TransitionPositionCommand(index=index, position=position)
        self.connection.mixer.send_commands([cmd])

    def on_next_clicked(self, widget, index, current):
        self.log_sw.debug('next', index, current)
        cmd = TransitionSettingsCommand(index=index, next_transition=current)
        self.connection.mixer.send_commands([cmd])

    def on_prev_trans_clicked(self, widget, index, enabled):
        cmd = TransitionPreviewCommand(index=index, enabled=enabled)
        self.connection.mixer.send_commands([cmd])

    def on_auto_rate_changed(self, widget, index, style, frames):
        if self.model_changing:
            return
        cmd = None
        # Send new rate to the mixer
        if style == 'mix':
            cmd = MixSettingsCommand(index=index, rate=frames)
        elif style == 'dip':
            cmd = DipSettingsCommand(index=index, rate=frames)
        elif style == 'wipe':
            cmd = WipeSettingsCommand(index=index, rate=frames)
        elif style == 'dve':
            cmd = DveSettingsCommand(index=index, rate=frames)
        if cmd is not None:
            self.connection.mixer.send_commands([cmd])

    def on_dip_source_changed(self, widget):
        if hasattr(widget, 'ignore_change') and widget.ignore_change or self.model_changing:
            return
        cmd = DipSettingsCommand(index=0, source=int(self.dip_source.get_active_id()))
        self.connection.mixer.send_commands([cmd])

    def on_aux_source_changed(self, widget):
        if hasattr(widget, 'ignore_change') and widget.ignore_change or self.model_changing:
            return

        self.routing.change(widget.index, int(widget.get_active_id()))

    def on_wipe_symmetry_adj_value_changed(self, widget, *args):
        if self.model_changing:
            return
        cmd = WipeSettingsCommand(index=0, symmetry=int(widget.get_value()))
        self.connection.mixer.send_commands([cmd])

    def on_wipe_x_adj_value_changed(self, widget, *args):
        if self.model_changing:
            return
        cmd = WipeSettingsCommand(index=0, positionx=int(widget.get_value()))
        self.connection.mixer.send_commands([cmd])

    def on_wipe_y_adj_value_changed(self, widget, *args):
        if self.model_changing:
            return
        cmd = WipeSettingsCommand(index=0, positiony=int(widget.get_value()))
        self.connection.mixer.send_commands([cmd])

    def on_wipe_width_adj_value_changed(self, widget, *args):
        if self.model_changing:
            return
        cmd = WipeSettingsCommand(index=0, width=int(widget.get_value()))
        self.connection.mixer.send_commands([cmd])

    def on_wipe_softness_adj_value_changed(self, widget, *args):
        if self.model_changing:
            return
        cmd = WipeSettingsCommand(index=0, softness=int(widget.get_value()))
        self.connection.mixer.send_commands([cmd])

    def on_wipe_fill_changed(self, widget, *args):
        if hasattr(widget, 'ignore_change') and widget.ignore_change or self.model_changing:
            return
        cmd = WipeSettingsCommand(index=0, source=int(widget.get_active_id()))
        self.connection.mixer.send_commands([cmd])
        self.focus_dummy.grab_focus()

    def on_wipe_flipflop_clicked(self, widget, *args):
        if self.model_changing:
            return
        state = widget.get_style_context().has_class('active')
        cmd = WipeSettingsCommand(index=0, flipflop=not state)
        self.connection.mixer.send_commands([cmd])

    def on_wipe_reverse_clicked(self, widget, *args):
        if self.model_changing:
            return
        state = widget.get_style_context().has_class('active')
        cmd = WipeSettingsCommand(index=0, reverse=not state)
        self.connection.mixer.send_commands([cmd])

    def on_ftb_afv_clicked(self, widget, *args):
        if self.mixer == 'atem':
            cmd = AudioMasterPropertiesCommand(afv=not widget.get_style_context().has_class('active'))
        else:
            cmd = FairlightMasterPropertiesCommand(afv=not widget.get_style_context().has_class('active'))
        self.connection.mixer.send_commands([cmd])

    def on_rate_focus(self, *args):
        self.disable_shortcuts = True

    def on_rate_unfocus(self, *args):
        self.disable_shortcuts = False

    def on_slider_held(self, *args):
        self.slider_held = True

    def on_slider_released(self, *args):
        self.slider_held = False

    def frames_to_time(self, frames):
        # WTF BMD
        if self.mode.rate < 29:
            transition_rate = 25
        elif self.mode.rate < 49:
            transition_rate = 30
        elif self.mode.rate < 59:
            transition_rate = 25
        else:
            transition_rate = 30

        if self.mode is None:
            return '00:00'

        seconds = frames // transition_rate
        frames = frames % transition_rate
        return '{:02d}:{:02d}'.format(int(seconds), int(frames))

    def time_to_frames(self, timestr):
        if self.mode.rate < 29:
            transition_rate = 25
        elif self.mode.rate < 49:
            transition_rate = 30
        elif self.mode.rate < 59:
            transition_rate = 25
        else:
            transition_rate = 30

        if self.mode is None:
            return 0

        part = timestr.split(':')
        if len(part) == 1:
            return int(part[0]) * transition_rate
        elif len(part) == 2:
            return (int(part[0]) * transition_rate) + int(part[1])

    @field('fade-to-black')
    def on_ftb_change(self, data):
        if data.index > len(self.me) - 1:
            self.log_sw.warning("Got FTB change for non-existing M/E {}".format(data.index + 1))
            return
        self.me[data.index].set_ftb_rate(data.rate)

    @field('transition-mix')
    def on_transition_mix_change(self, data):
        if data.index > len(self.me) - 1:
            self.log_sw.warning("Got transition mix change for non-existing M/E {}".format(data.index + 1))
            return

        label = self.frames_to_time(data.rate)
        self.mix_rate.set_text(label)
        self.me[data.index].set_auto_rate('mix', data.rate)

    @field('transition-dip')
    def on_transition_dip_change(self, data):
        if data.index > len(self.me) - 1:
            self.log_sw.warning("Got transition dip change for non-existing M/E {}".format(data.index + 1))
            return

        label = self.frames_to_time(data.rate)
        self.dip_rate.set_text(label)
        self.dip_source.ignore_change = True
        self.dip_source.set_active_id(str(data.source))
        self.dip_source.ignore_change = False
        self.me[data.index].set_auto_rate('dip', data.rate)

    @field('transition-wipe')
    def on_transition_wipe_change(self, data):
        if data.index > len(self.me) - 1:
            self.log_sw.warning("Got transition wipe change for non-existing M/E {}".format(data.index + 1))
            return

        label = self.frames_to_time(data.rate)
        self.wipe_rate.set_text(label)
        self.me[data.index].set_auto_rate('wipe', data.rate)

        for style, button in enumerate(self.wipe_style):
            self.set_class(button, 'stylebtn', True)
            self.set_class(button, 'active', style == data.pattern)

        if not self.slider_held:
            self.model_changing = True
            self.wipe_symmetry_adj.set_value(data.symmetry)
            self.wipe_x_adj.set_value(data.positionx)
            self.wipe_y_adj.set_value(data.positiony)
            self.wipe_width_adj.set_value(data.width)
            self.wipe_softness_adj.set_value(data.softness)
            self.model_changing = False

        self.model_changing = True
        self.wipe_fill.set_active_id(str(data.source))
        self.model_changing = False

        self.wipe_fill.set_sensitive(data.width > 0)
        self.set_class(self.wipe_reverse, 'active', data.reverse)
        self.set_class(self.wipe_flipflop, 'active', data.flipflop)

    @field('transition-dve')
    def on_transition_dve_change(self, data):
        if data.index > len(self.me) - 1:
            self.log_sw.warning("Got transition dve change for non-existing M/E {}".format(data.index + 1))
            return

        for button in self.dve_style_push + self.dve_style_squeeze:
            self.set_class(button, 'stylebtn', True)
            self.set_class(button, 'active', data.style == button.styleindex)

        self.set_class(self.dve_reverse, 'active', data.reverse)
        self.set_class(self.dve_flipflop, 'active', data.flipflop)

        label = self.frames_to_time(data.rate)
        self.dve_rate.set_text(label)
        self.me[data.index].set_auto_rate('dve', data.rate)

    def _remap(self, value, old_min, old_max, new_min, new_max):
        return ((value - old_min) / (old_max - old_min)) * (new_max - new_min) + new_min

    @field('dkey-properties')
    def on_dsk_change(self, data):
        self.me[0].set_dsk(data)
        top = 9000 - data.top
        bottom = data.bottom + 9000
        left = data.left + 16000
        right = 16000 - data.right

        region = self.layout[0].get(LayoutView.LAYER_DSK, data.index)
        region.set_mask(top, bottom, left, right)

    @field('dkey-state')
    def on_dsk_state_change(self, data):
        self.me[0].set_dsk_state(data)
        region = self.layout[0].get(LayoutView.LAYER_DSK, data.index)
        region.set_tally(data.on_air)

    @field('topology')
    def on_topology_change(self, data: TopologyField):
        # Create the M/E units
        for i in range(0, data.me_units - len(self.me)):
            self.add_mixeffect()

        # Color generators
        colorpanel = ColorPanel(self.connection.mixer)
        colorpanel.window = self.window
        colorpanel.application = self.application
        self.apply_css(colorpanel, self.provider)
        self.palette_top.add(colorpanel)

        # Downstream keyer count, only available on M/E 1
        self.me[0].set_topology(data)
        self.apply_css(self.me[0], self.provider)

        from gtk_switcher.panel_downstreamkeyer import DownstreamKeyerPanel
        for widget in self.downstream_keyers:
            self.downstream_keyers.remove(widget)

        for i in range(0, data.downstream_keyers):
            panel = DownstreamKeyerPanel(self.connection.mixer, i, self.model_me1_fill, self.model_key)
            self.apply_css(panel, self.provider)
            self.downstream_keyers.pack_start(panel, False, True, 0)

            # Add the DSK to the layout editor of M/E 1
            region = self.layout[0].get(LayoutView.LAYER_DSK, i)
            region.set_region(0, 0, 16, 9)
            region.set_mask(0, 0, 0, 0)
        self.downstream_keyers.show_all()

        # Media players
        for i in range(0, data.mediaplayers):
            panel = MediaPlayerPanel(i, model_media=self.model_media, connection=self.connection.mixer)
            self.switcher_mediaplayers.add(panel)
        self.media_create_mediaplayers(data.mediaplayers)
        self.switcher_mediaplayers.show_all()

        # Supersource
        for i in range(0, data.supersources):
            sspanel = SupersourcePanel(self.connection.mixer, i, self.model_supersource_box, self.model_supersource_art)
            sspanel.window = self.window
            sspanel.application = self.application
            self.apply_css(sspanel, self.provider)
            self.palette_top.add(sspanel)

    def on_dsk_tie_clicked(self, widget, index, dsk, enabled):
        cmd = DkeyTieCommand(index=dsk, tie=enabled)
        self.connection.mixer.send_commands([cmd])

    def on_dsk_onair_clicked(self, widget, index, dsk, enabled):
        cmd = DkeyOnairCommand(index=dsk, on_air=enabled)
        self.connection.mixer.send_commands([cmd])

    def on_dsk_rate_activate(self, widget, index, dsk, frames):
        cmd = DkeyRateCommand(index=dsk, rate=frames)
        self.connection.mixer.send_commands([cmd])

    def on_dsk_auto_clicked(self, widget, index, dsk):
        cmd = DkeyAutoCommand(index=dsk)
        self.connection.mixer.send_commands([cmd])

    @field('mixer-effect-config')
    def on_mixer_effect_config_change(self, data):
        if data.index > len(self.me) - 1:
            self.log_sw.warning("Got _MeC for non-existing M/E {}".format(data.index + 1))
            return

        if data.index not in self.usks:
            self.usks[data.index] = {}

        self.me[data.index].set_config(data)

        from gtk_switcher.upstreamkey import UpstreamKeyer

        usk_count = len(self.usks[data.index])
        if data.keyers > usk_count:
            add = data.keyers - usk_count

            for i in range(0, add):
                exp = Gtk.Expander()
                exp.get_style_context().add_class('bmdgroup')
                label_text = _("Upstream keyer {}").format(len(self.usks[data.index]) + 1)
                if data.index != 0:
                    label_text += ' [M/E {}]'.format(data.index + 1)
                frame_label = Gtk.Label(label_text)
                frame_label.get_style_context().add_class("heading")
                exp.set_label_widget(frame_label)
                exp.set_expanded(True)
                usk = UpstreamKeyer(index=data.index, keyer=len(self.usks[data.index]), connection=self.connection)
                self.usks[data.index][usk.keyer] = usk
                self.has_models.append(usk)
                exp.add(usk)
                self.apply_css(usk, self.provider)
                usk.set_fill_model(self.model_me1_fill)
                usk.set_key_model(self.model_key)
                self.upstream_keyers.pack_start(exp, False, True, 0)

            self.upstream_keyers.show_all()

    @field('fade-to-black-state')
    def on_ftb_state_change(self, data):
        if data.index > len(self.me) - 1:
            self.log_sw.warning("Got FTB state change for non-existing M/E {}".format(data.index + 1))
            return

        self.me[data.index].set_ftb_state(data.done, data.transitioning)

    @field('key-on-air')
    def on_key_on_air_change(self, data):
        if data.index > len(self.me) - 1:
            self.log_sw.warning("Got key-on-air for non-existant keyer {} M/E {}".format(data.keyer, data.index + 1))
            return
        self.me[data.index].set_key_on_air(data)
        self.layout[data.index].get(LayoutView.LAYER_USK, data.keyer).set_tally(data.enabled)

    @field('transition-preview')
    def on_transition_preview_change(self, data):
        if data.index > len(self.me) - 1:
            self.log_sw.warning("Got transition preview change for non-existing M/E {}".format(data.index + 1))
            return
        self.me[data.index].set_preview_transition(data.enabled)

    @field('transition-settings')
    def on_transition_settings_change(self, data):
        if data.index > len(self.me) - 1:
            self.log_sw.warning("Got transition settings change for non-existing M/E {}".format(data.index + 1))
            return
        self.me[data.index].set_transition_settings(data)

    @field('transition-position')
    def on_transition_position_change(self, data):
        if data.index > len(self.me) - 1:
            self.log_sw.warning("Got transition position change for non-existing M/E {}".format(data.index + 1))
            return
        self.me[data.index].set_transition_progress(data)

    def on_onair_clicked(self, widget, index, keyer, enabled):
        cmd = KeyOnAirCommand(index=index, keyer=keyer, enabled=enabled)
        self.connection.mixer.send_commands([cmd])

    @field('key-properties-base')
    def on_key_properties_base_change(self, data):
        if data.index not in self.usks:
            self.log_sw.warning(
                "Got key-properties-base for non-existant M/E {}".format(data.index + 1))
            return
        if data.keyer not in self.usks[data.index]:
            self.log_sw.warning(
                "Got key-properties-base for non-existant keyer {} M/E {}".format(data.keyer, data.index + 1))
            return
        self.usks[data.index][data.keyer].on_key_properties_base_change(data)

    @field('key-properties-luma')
    def on_key_properties_luma_change(self, data):
        if data.index not in self.usks:
            self.log_sw.warning(
                "Got key-properties-luma for non-existant M/E {}".format(data.index + 1))
            return
        if data.keyer not in self.usks[data.index]:
            self.log_sw.warning(
                "Got key-properties-luma for non-existant keyer {} M/E {}".format(data.keyer, data.index + 1))
            return
        self.usks[data.index][data.keyer].on_key_properties_luma_change(data)

    @field('key-properties-dve')
    def on_key_properties_dve_change(self, data):
        if data.index not in self.usks:
            self.log_sw.warning(
                "Got key-properties-dve for non-existant M/E {}".format(data.index + 1))
            return
        if data.keyer not in self.usks[data.index]:
            self.log_sw.warning(
                "Got key-properties-dve for non-existant keyer {} M/E {}".format(data.keyer, data.index + 1))
            return

        self.usks[data.index][data.keyer].on_key_properties_dve_change(data)
        width = 16.0 * data.size_x / 1000
        height = 9.0 * data.size_y / 1000

        region = self.layout[data.index].get(LayoutView.LAYER_USK, data.keyer)
        region.set_region(data.pos_x / 1000, data.pos_y / 1000, width, height)
        region.set_mask(data.mask_top, data.mask_bottom, data.mask_left, data.mask_right)

    @field('key-properties-advanced-chroma')
    def on_key_properties_advanced_chroma_change(self, data):
        if data.index not in self.usks:
            self.log_sw.warning(
                "Got KACk for non-existant M/E {}".format(data.index + 1))
            return
        if data.keyer not in self.usks[data.index]:
            self.log_sw.warning(
                "Got KACk for non-existant keyer {} M/E {}".format(data.keyer, data.index + 1))
            return
        self.usks[data.index][data.keyer].on_advanced_chroma_change(data)

    @field('key-properties-advanced-chroma-colorpicker')
    def on_key_properties_advanced_chroma_colorpicker_change(self, data):
        if data.index not in self.usks:
            self.log_sw.warning(
                "Got KACC for non-existant M/E {}".format(data.index + 1))
            return
        if data.keyer not in self.usks[data.index]:
            self.log_sw.warning(
                "Got KACC for non-existant keyer {} M/E {}".format(data.keyer, data.index + 1))
            return

        self.usks[data.index][data.keyer].on_chroma_picker_change(data)

        size = data.size / 1000

        region = self.layout[data.index].get(LayoutView.LAYER_ACK, data.keyer)
        region.set_region(data.x / 1000, data.y / 1000, size, size)
        region.set_tally(data.preview)
        region.set_visible(data.cursor)

    @field('program-bus-input')
    def on_program_input_change(self, data):
        if data.index > len(self.me) - 1:
            self.log_sw.warning("Got program input change for non-existing M/E {}".format(data.index + 1))
            return
        self.me[data.index].program_input_change(data)

    @field('preview-bus-input')
    def on_preview_input_change(self, data):
        if data.index > len(self.me) - 1:
            self.log_sw.warning("Got preview input change for non-existing M/E {}".format(data.index + 1))
            return
        self.me[data.index].preview_input_change(data)

    def on_me_program_changed(self, widget, index, source):
        cmd = ProgramInputCommand(index=index, source=source)
        self.connection.mixer.send_commands([cmd])

    def on_me_preview_changed(self, widget, index, source):
        cmd = PreviewInputCommand(index=index, source=source)
        self.connection.mixer.send_commands([cmd])

    @field('recording-settings')
    def on_stream_recording_setting_change(self, data: RecordingSettingsField):
        self.expander_stream_recorder.show()
        self.model_changing = True
        self.stream_recorder_filename.set_text(data.filename)
        self.stream_recorder_disk = [data.disk1, data.disk2]

        self.stream_recorder_disk1.set_active_id(str(data.disk1))
        self.stream_recorder_disk2.set_active_id(str(data.disk2))
        self.stream_recorder_trigger_cameras.set_active(data.record_in_cameras)
        self.on_update_recording_buttons()
        self.model_changing = False

    @field('recording-disk')
    def on_stream_recording_disks_change(self, data):
        self.model_changing = True
        if data.is_deleted:
            del self.disks[data.index]
        else:
            self.disks[data.index] = data
            for i, row in enumerate(self.model_disks):
                if row[0] == str(data.index):
                    self.model_disks[i][1] = data.volumename
                    break
            else:
                self.model_disks.append([str(data.index), data.volumename])

        self.stream_recorder_disk1.set_active_id(str(self.stream_recorder_disk[0]))
        self.stream_recorder_disk2.set_active_id(str(self.stream_recorder_disk[1]))
        self.stream_recorder_disk1.set_sensitive(self.stream_recorder_disk[0] is not None)
        self.stream_recorder_disk2.set_sensitive(self.stream_recorder_disk[1] is not None)
        self.on_update_recording_buttons()
        self.model_changing = False

    @field('recording-status')
    def on_stream_recording_status_change(self, data):
        self.on_update_recording_buttons()

        status = 'STOP'
        active = False
        if data.has_dropped:
            status = 'DROP'
            active = True
        elif data.is_stopping:
            status = 'STOPPING'
        elif data.disk_full:
            status = 'DISK FULL'
        elif data.is_recording:
            status = 'REC'
            active = True

        if active != self.stream_recorder_active:
            if active:
                self.stream_recorder_start_time = datetime.now().timestamp()

        self.set_class(self.headerbar, 'recording', active)
        self.stream_recorder_active = active

        self.stream_recorder_status.set_text(status)
        self.set_class(self.stream_recorder_status, 'program', active)
        self.set_class(self.stream_recorder_clock, 'program', active)

    @field('recording-duration')
    def on_stream_recording_duration_change(self, data):
        # This does not get called nearly often enough
        self.stream_recorder_clock.set_text(f'{data.hours}:{data.minutes}:{data.seconds}')
        seconds = data.seconds + (60 * data.minutes) + (60 * 60 * data.hours)
        self.stream_recorder_start_time = datetime.now().timestamp() - seconds

    def on_clock_stream_recorder(self):
        if self.stream_recorder_active:
            length = timedelta(seconds=int(datetime.now().timestamp() - self.stream_recorder_start_time))
            self.stream_recorder_clock.set_text(str(length))

    def on_clock_stream_live(self):
        if self.stream_live_active:
            length = timedelta(seconds=int(datetime.now().timestamp() - self.stream_live_start_time))
            self.stream_live_clock.set_text(str(length))

    def on_update_recording_buttons(self):
        has_usable_disks = False
        for index in self.disks:
            if self.disks[index].is_ready:
                has_usable_disks = True
        has_multi_ready_disks = self.stream_recorder_disk[0] is not None and self.stream_recorder_disk[1] is not None

        if 'recording-status' not in self.connection.mixer.mixerstate:
            return

        status = self.connection.mixer.mixerstate['recording-status']
        is_recording = status.is_recording

        self.stream_recorder_start.set_sensitive(has_usable_disks and not is_recording)
        self.stream_recorder_stop.set_sensitive(is_recording)
        self.stream_recorder_switch.set_sensitive(has_multi_ready_disks and is_recording)

        self.set_class(self.stream_recorder_disk1_label, 'program',
                       self.stream_recorder_disk[0] is not None
                       and self.stream_recorder_disk[0] in self.disks
                       and self.disks[self.stream_recorder_disk[0]].is_recording)
        self.set_class(self.stream_recorder_disk2_label, 'program',
                       self.stream_recorder_disk[1] is not None
                       and self.stream_recorder_disk[1] in self.disks
                       and self.disks[self.stream_recorder_disk[1]].is_recording)

    def on_stream_recorder_start_clicked(self, widget, *args):
        cmd = RecorderStatusCommand(recording=True)
        self.connection.mixer.send_commands([cmd])

    def on_stream_recorder_stop_clicked(self, widget, *args):
        cmd = RecorderStatusCommand(recording=False)
        self.connection.mixer.send_commands([cmd])

    def on_stream_recorder_trigger_cameras_clicked(self, widget, *args):
        cmd = RecordingSettingsSetCommand(record_in_camera=widget.get_active())
        self.connection.mixer.send_commands([cmd])

    def on_aux_output_source_change(self, source):
        self.routing.aux_changed(source)
        aux = self.routing.aux_index_to_port_index(source.index)
        if aux not in self.aux:
            return

        self.aux[aux].ignore_change = True
        self.aux[aux].set_active_id(str(source.source))
        self.aux[aux].ignore_change = False

        for me in self.me:
            if not hasattr(me, 'category'):
                continue
            if me.index == aux:
                me.source_change(source.source)

    def on_input_layout_change(self, changed_input):
        inputs = self.connection.mixer.mixerstate['input-properties']
        external = []
        colors = []
        mp = []
        black = None
        bars = None
        extra1 = []
        extra2 = []

        extra1_me1 = []
        extra2_me1 = []
        extra1_me2 = []
        extra2_me2 = []

        # Clear the combobox models
        self.model_changing = True
        for i in self.has_models:
            i.model_changing = True
        self.model_me1_fill.clear()
        self.model_aux.clear()
        self.model_key.clear()

        for i in list(inputs.values()):
            if i.port_type == InputPropertiesField.PORT_EXTERNAL:
                external.append(i)
            if i.port_type == InputPropertiesField.PORT_COLOR:
                colors.append(i)
            if i.port_type == InputPropertiesField.PORT_MEDIAPLAYER and i.available_me1:
                mp.append(i)
            if i.port_type == InputPropertiesField.PORT_BLACK:
                black = i
            if i.port_type == InputPropertiesField.PORT_BARS:
                bars = i
            if i.port_type == InputPropertiesField.PORT_SUPERSOURCE:
                extra2.append(i)
            if i.port_type == InputPropertiesField.PORT_ME_OUTPUT:
                if i.available_me1:
                    extra1_me1.append(i)
                if i.available_me2:
                    extra1_me2.append(i)

            if i.available_me1:
                self.model_me1_fill.append([str(i.index), i.name])
            if i.available_key_source:
                self.model_key.append([str(i.index), i.name])

            if i.available_aux:
                self.model_aux.append([str(i.index), i.name])

            if i.available_supersource_box:
                self.model_supersource_box.append([str(i.index), i.name])
            if i.available_supersource_art:
                self.model_supersource_art.append([str(i.index), i.name])

            if i.port_type == InputPropertiesField.PORT_AUX_OUTPUT:
                self.routing.add_output(i)
                aux_id = i.index
                if aux_id not in self.aux:
                    # Create AUX combobox in palette
                    self.aux[aux_id] = Gtk.ComboBox.new_with_model(self.model_aux)
                    self.aux[aux_id].set_entry_text_column(1)
                    self.aux[aux_id].set_id_column(0)
                    self.aux[aux_id].index = aux_id
                    self.aux[aux_id].connect('changed', self.on_aux_source_changed)
                    renderer = Gtk.CellRendererText()
                    self.aux[aux_id].pack_start(renderer, True)
                    self.aux[aux_id].add_attribute(renderer, "text", 1)
                    self.grid_aux.attach(self.aux[aux_id], 1, aux_id, 1, 1)

                    # Create options dropdown after the combobox
                    aux_me = Gtk.CheckButton.new_with_label(_("Show as bus"))
                    aux_me.index = i.index
                    aux_me.connect('toggled', self.on_aux_me_enable_toggled)

                    aux_follow_mon = Gtk.CheckButton.new_with_label(_("Follow audio monitor bus"))
                    aux_follow_mon.index = i.index
                    aux_follow_mon.connect('toggled', self.on_aux_me_follow_toggled)

                    aux_btn = Gtk.MenuButton()
                    hamburger = Gtk.Image.new_from_icon_name('open-menu-symbolic', Gtk.IconSize.BUTTON)
                    aux_btn.set_image(hamburger)

                    popover = Gtk.PopoverMenu()
                    aux_btn.set_popover(popover)
                    popover_box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
                    popover_box.set_margin_top(8)
                    popover_box.set_margin_bottom(8)
                    popover_box.set_margin_start(8)
                    popover_box.set_margin_end(8)
                    popover_box.add(aux_me)
                    popover_box.add(aux_follow_mon)
                    popover.add(popover_box)
                    popover_box.show_all()
                    self.grid_aux.attach(aux_btn, 2, aux_id, 1, 1)

                    aux_label = Gtk.Label(label=i.name)
                    aux_label.get_style_context().add_class('dim-label')
                    self.grid_aux.attach(aux_label, 0, aux_id, 1, 1)
                    self.grid_aux.show_all()

        row1_ext = external
        row2_ext = [None] * len(external)
        if len(external) > 4:
            num = len(external) // 2
            row1_ext = external[0:num]
            row2_ext = external[num:] + [None] * ((2 * num) - len(external))

        row1 = row1_ext + [None, black, None] + colors + extra1
        row2 = row2_ext + [None, bars, None] + mp + extra2

        buttons = [row1, row2]

        for me in self.me:
            me_btns = [buttons[0][:], buttons[1][:]]
            if me.index == 0:
                me_btns[0] += extra1_me1
                me_btns[1] += extra2_me1
            elif me.index == 1:
                me_btns[0] += extra1_me2
                me_btns[1] += extra2_me2
            me.set_inputs(me_btns)
            self.apply_css(me, self.provider)

        self.model_changing = False
        for i in self.has_models:
            i.model_changing = False

    def on_aux_me_follow_toggled(self, widget):
        aux_id = widget.index - 8001
        if widget.get_active():
            self.aux_follow_audio.add(aux_id)
        else:
            self.aux_follow_audio.remove(aux_id)

    def on_aux_me_enable_toggled(self, widget):
        from gtk_switcher.mixeffect_aux import AuxMixEffectBlock
        if widget.get_active():
            inputs = self.connection.mixer.mixerstate['input-properties']
            auxsrcs = self.connection.mixer.mixerstate['aux-output-source']
            webcam = widget.index == 8200
            aux = self.routing.port_index_to_aux_index(widget.index)
            auxsrc = auxsrcs[aux]
            external = []
            output = []
            passthrough = []
            special = []
            black = []
            bars = []
            for i in inputs.values():
                if i.available_aux and not webcam or webcam and i.available_usb:
                    if i.port_type == InputPropertiesField.PORT_EXTERNAL:
                        external.append(i)
                    elif i.port_type == InputPropertiesField.PORT_ME_OUTPUT:
                        output.append(i)
                    elif i.port_type == InputPropertiesField.PORT_PASSTHROUGH:
                        passthrough.append(i)
                    elif i.port_type == InputPropertiesField.PORT_BLACK:
                        black.append(None)
                        black.append(i)
                    elif i.port_type == InputPropertiesField.PORT_BARS:
                        bars.append(None)
                        bars.append(i)
                    else:
                        special.append(i)

            row1_ext = external
            row2_ext = [None] * len(external)
            if len(external) > 6:
                num = len(external) // 2
                row1_ext = external[0:num]
                row2_ext = external[num:] + [None] * ((2 * num) - len(external))

                row1 = row1_ext + black + [None] + output + passthrough
                row2 = row2_ext + bars + [None] + special
            else:
                row1 = row1_ext + [None] + passthrough + black + bars + [None] + output + special
                row2 = row2_ext

            name = inputs[widget.index].name
            aux_me = AuxMixEffectBlock(widget.index, name)
            aux_me.set_inputs([row1, row2])
            aux_me.source_change(auxsrc.source)
            aux_me.connect('source-changed', self.on_aux_me_source_changed)
            aux_me.index = widget.index
            aux_me.category = 'aux'
            self.apply_css(aux_me, self.provider)
            self.me.append(aux_me)
            self.main_blocks.add(aux_me)
        else:
            for me in self.me:
                if not hasattr(me, 'category'):
                    continue
                if me.category != 'aux':
                    continue
                if me.index == widget.index:
                    self.me.remove(me)
                    me.destroy()

    def on_aux_me_source_changed(self, widget, aux, source):
        self.routing.change(aux, source)

    @field('macro-properties')
    def on_macro_properties_change(self, data):
        # Clear the macro flow container
        for widget in self.macro_flow:
            self.macro_flow.remove(widget)

        # Create new buttons
        macros = self.connection.mixer.mixerstate['macro-properties']
        for index in macros:
            macro = macros[index]
            if macro.is_used:
                button = Gtk.Button(macro.name.decode())
                button.index = index
                button.get_style_context().add_class('bmdbtn')
                button.get_style_context().add_class('macro')
                button.connect('button-press-event', self.on_macro_context)
                self.macro_flow.add(button)
        self.macro_flow.show_all()

    def on_macro_run(self, widget):
        index = widget.index
        cmd = MacroActionCommand(MacroActionCommand.ACTION_RUN, index=index)
        self.connection.mixer.send_commands([cmd])

    def on_macro_delete(self, widget):
        index = widget.index
        cmd = MacroActionCommand(MacroActionCommand.ACTION_DELETE, index=index)
        self.connection.mixer.send_commands([cmd])

    def on_macro_context(self, widget, event, *args):
        if event.button == 1:
            self.on_macro_run(widget)

        if event.button != 3:
            return

        self.menu = Gtk.Menu()

        run_item = Gtk.MenuItem(_("Run macro"))
        run_item.index = widget.index
        run_item.connect('activate', self.on_macro_run)
        self.menu.append(run_item)

        edit_item = Gtk.MenuItem(_("Edit macro"))
        edit_item.index = widget.index
        edit_item.connect('activate', self.on_macro_edit)
        self.menu.append(edit_item)

        edit_item = Gtk.MenuItem(_("Delete macro"))
        edit_item.index = widget.index
        edit_item.connect('activate', self.on_macro_delete)
        self.menu.append(edit_item)

        self.menu.show_all()
        self.menu.popup(None, None, None, None, 0, Gtk.get_current_event_time())

    def on_macro_edit(self, widget):
        self.macro_edit = True
        self.connection.mixer.download(0xffff, widget.index)

    def on_macro_create_clicked(self, widget):
        name = self.macro_name.get_text()
        if name.strip() == '':
            return

        # Find a free macro slot for recording
        slot_index = 0
        for slot in self.connection.mixer.mixerstate['macro-properties']:
            mp = self.connection.mixer.mixerstate['macro-properties'][slot]
            if not mp.is_used:
                slot_index = slot
                break
        else:
            self.log_sw.error("Could not find a free macro slot to record in")
            return

        self.log_sw.info(f'Creating macro "{name}" in slot {slot_index}')

        cmd = MacroRecordCommand(slot_index, name, '')
        self.connection.mixer.send_commands([cmd])

    def on_macro_status_stop_clicked(self, widget):
        cmd = MacroActionCommand(MacroActionCommand.ACTION_STOP_RECORD)
        self.connection.mixer.send_commands([cmd])

    def bps_to_human(self, bps):
        if bps < 1000:
            return str(bps) + 'b'
        elif bps < 1000 * 1000:
            val = bps / 1000
            return f'{val:g}k'
        else:
            val = bps / 1000 / 1000
            return f'{val:g}M'

    @field('streaming-service')
    def on_streaming_service_change(self, data):
        self.expander_livestream.show()
        self.expander_encoder.show()
        self.video_rate_min.set_text(self.bps_to_human(data.min))
        self.video_rate_max.set_text(self.bps_to_human(data.max))

        self.stream_live_platform.set_text(data.name)
        self.stream_live_server.set_text(data.url)
        self.stream_live_key.set_text(data.key)

    @field('streaming-audio-bitrate')
    def on_streaming_audio_bitrate_change(self, data):
        self.expander_encoder.show()
        self.audio_rate_min.set_text(self.bps_to_human(data.min))
        self.audio_rate_max.set_text(self.bps_to_human(data.max))

    @field('streaming-status')
    def on_streaming_status_change(self, data):
        starting = data.status == 2
        active = data.status == 4
        self.set_class(self.headerbar, 'streaming', active)
        if active:
            self.live_stats.show()
        else:
            self.live_stats.hide()

        self.stream_live_start.set_sensitive(not starting and not active)
        self.stream_live_stop.set_sensitive(starting or active)

        status = {
            1: (gettext.pgettext("livestream", "OFF"), False, False),
            2: (gettext.pgettext("livestream", "starting..."), True, False),
            4: (gettext.pgettext("livestream", "ON AIR"), False, True),
            34: (gettext.pgettext("livestream", "stopping..."), True, False),
            36: (gettext.pgettext("livestream", "stopping..."), True, False),
        }
        if data.status in status:
            self.stream_live_status.set_text(status[data.status][0])
            self.set_class(self.stream_live_status, 'active', status[data.status][1])
            self.set_class(self.stream_live_status, 'program', status[data.status][2])

        if active != self.stream_live_active:
            if active:
                self.stream_live_start_time = datetime.now().timestamp()

        self.stream_live_active = active

    @field('streaming-stats')
    def on_streaming_stats_change(self, data):
        self.live_stats.set_text('{:.2f} Mbps'.format(data.bitrate / 1000 / 1000))

    def create_livestream_presets(self):
        hardcoded = gtk_switcher.stream_data.services

        menu = Gio.Menu()
        for provider in hardcoded:
            section = Gio.Menu()
            for preset in hardcoded[provider]:
                item = Gio.MenuItem()
                item.set_label(preset.name)
                item.service = preset
                item.set_action_and_target_value('app.streampreset', preset.variant())
                section.append_item(item)

            menu.append_section(provider, section)
        self.stream_presets.set_menu_model(menu)

    def load_livestream_preset(self, action, data):
        data = dict(data)
        cmd = StreamingServiceSetCommand(name=data['name'], url=data['url'])
        self.connection.mixer.send_commands([cmd])

    def on_stream_live_platform_activate(self, widget, *args):
        cmd = StreamingServiceSetCommand(name=widget.get_text())
        self.connection.mixer.send_commands([cmd])

    def on_stream_live_server_activate(self, widget, *args):
        cmd = StreamingServiceSetCommand(url=widget.get_text())
        self.connection.mixer.send_commands([cmd])

    def on_stream_live_key_activate(self, widget, *args):
        cmd = StreamingServiceSetCommand(key=widget.get_text())
        self.connection.mixer.send_commands([cmd])

    def on_stream_live_start_clicked(self, *args):
        cmd = StreamingStatusSetCommand(True)
        self.connection.mixer.send_commands([cmd])

    def on_stream_live_stop_clicked(self, *args):
        cmd = StreamingStatusSetCommand(False)
        self.connection.mixer.send_commands([cmd])

    def on_stream_recorder_disk1_changed(self, widget, *args):
        if self.model_changing:
            return
        disk_id = int(widget.get_active_id())
        cmd = RecordingSettingsSetCommand(disk1=disk_id)
        self.connection.mixer.send_commands([cmd])

    def on_stream_recorder_disk2_changed(self, widget, *args):
        if self.model_changing:
            return
        disk_id = int(widget.get_active_id())
        cmd = RecordingSettingsSetCommand(disk2=disk_id)
        self.connection.mixer.send_commands([cmd])

    def on_stream_recorder_filename_activate(self, widget, *args):
        cmd = RecordingSettingsSetCommand(filename=widget.get_text())
        self.connection.mixer.send_commands([cmd])

    def human_to_bps(self, human):
        human = human.lower()
        human = human.replace('ps', '')

        if human.isnumeric():
            num = human
            unit = ''
        else:
            if not ' ' in human:
                human = re.sub(r'([a-z]+)', r' \1', human)
            num, unit = [string.strip() for string in human.split()]

        mult = 1
        if unit.startswith('k'):
            mult = 1000
        elif unit.startswith('m'):
            mult = 1000 * 1000
        return int(float(num) * mult)

    def on_audio_bitrate_activate(self, widget, *args):
        rate_min = self.human_to_bps(self.audio_rate_min.get_text())
        rate_max = self.human_to_bps(self.audio_rate_max.get_text())
        if rate_max < rate_min:
            rate_max = rate_min
        cmd = StreamingAudioBitrateCommand(rate_min, rate_max)
        self.connection.mixer.send_commands([cmd])

    def on_video_bitrate_activate(self, widget, *args):
        rate_min = self.human_to_bps(self.video_rate_min.get_text())
        rate_max = self.human_to_bps(self.video_rate_max.get_text())
        if rate_max < rate_min:
            rate_max = rate_min
        cmd = StreamingServiceSetCommand(bitrate_min=rate_min, bitrate_max=rate_max)
        self.connection.mixer.send_commands([cmd])

    def on_fold_folded(self, widget, *args):
        self.flaptoggle.set_visible(self.flap.get_folded())

    def on_presets_clicked(self, widget, *args):
        self.preset_context = widget.get_name()

    @field('supersource-box-properties')
    def on_supersource_box_change(self, data):
        x = data.x / 100
        y = data.y / 100
        width = data.size / 110 * 16 / 9
        height = data.size / 110

        region = self.layout[0].get(LayoutView.LAYER_SSB, data.box)
        region.set_region(x, y, width, height)
        region.set_mask(data.mask_top, data.mask_bottom, data.mask_left, data.mask_right)



# Copyright 2021 - 2022, Martijn Braam and the OpenAtem contributors
# SPDX-License-Identifier: LGPL-3.0-only
from unittest import TestCase

from pyatem.hexdump import hexdump
from pyatem.media import rle_decode, rle_encode, atem_to_image, image_to_atem, rgb_to_atem
from pyatem.mediaconvert import atem_to_rgb


class Test(TestCase):
    FRAME_1080_RED = b'\xfe\xfe\xfe\xfe\xfe\xfe\xfe\xfe\x00\x00\x00\x00\x00\x0f\xd2\x00:\x96d\xfa:\x9e\xfc\xfa'
    FRAME_1080_RED_PIXEL = b':\x96d\xfa:\x9e\xfc\xfa'
    RED_PIXEL_8888 = b'\xff\0\0\xff'

    def test_rle_decode(self):
        decompressed = rle_decode(self.FRAME_1080_RED)
        for index in range(0, len(decompressed), 8):
            pixel = decompressed[index:index + 8]
            self.assertEqual(self.FRAME_1080_RED_PIXEL, pixel, msg='index {}'.format(index))

    def test_atem_to_image(self):
        decompressed = rle_decode(self.FRAME_1080_RED)
        result = atem_to_rgb(decompressed, 1920, 1080)
        for index in range(0, len(result), 4):
            pixel = result[index:index + 4]
            self.assertEqual(self.RED_PIXEL_8888, pixel, msg='index {}'.format(index))

    def test_image_to_atem(self):
        frame = atem_to_image(self.FRAME_1080_RED, 1920, 1080)
        result = rgb_to_atem(frame, 1920, 1080)
        compressed = rle_encode(result)
        hexdump(compressed)
        test = atem_to_rgb(result, 1920, 1080)

    def test_rle_decode_solid_color(self):
        result = rle_decode(self.FRAME_1080_RED)
        for index in range(0, len(result) - 8, 8):
            pixel = result[index:index + 8]
            self.assertEqual(self.FRAME_1080_RED_PIXEL, pixel)

    def test_rle_encode_solid_color(self):
        testframe = rle_decode(self.FRAME_1080_RED)
        compressed = rle_encode(testframe)
        decompressed = rle_decode(compressed)
        self.assertEqual(testframe, decompressed)



# Copyright 2021 - 2022, Martijn Braam and the OpenAtem contributors
# SPDX-License-Identifier: GPL-3.0-only
import gi

from gtk_switcher.colorwheel import ColorWheelWidget
from pyatem.field import InputPropertiesField

gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, GLib, GObject, Gio, Gdk

gi.require_version('Handy', '1')
from gi.repository import Handy


class CameraPage:
    def __init__(self, builder):
        self.camera_box = builder.get_object('camera_box')

    def on_camera_layout_change(self, data):
        inputs = self.connection.mixer.mixerstate['input-properties']

        for child in self.camera_box:
            child.destroy()

        for i in inputs.values():
            if i.port_type == InputPropertiesField.PORT_EXTERNAL:
                box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
                box.set_spacing(8)
                label = Gtk.Label(i.name)
                label.get_style_context().add_class('heading')
                box.pack_start(label, False, True, 0)
                camera_frame = Gtk.Frame()
                camera_frame.get_style_context().add_class('view')
                box.pack_start(camera_frame, True, True, 0)
                self.camera_box.pack_start(box, True, True, 0)

                grid = Gtk.Grid()
                grid.set_row_spacing(8)
                grid.set_column_spacing(8)
                camera_frame.add(grid)

        self.camera_box.show_all()



"""
Generate a wireshark dissector based on the protocol information in the pyatem module
Based on the atem_dissector.lua wireshark plugin from https://github.com/peschuster/wireshark-atem-dissector
"""
import os
import pyatem.protocol
import pyatem.field


def make_pf_fields():
    result = "local pf_fields = {}\nlocal VALS = {}\n"
    for name, cls in pyatem.field.__dict__.items():
        if not name.endswith('Field'):
            continue
        cmd = cls.CODE

        in_field_table = False
        parsed_field_table = False
        has_header = False
        for line in cls.__doc__.splitlines(keepends=False):
            line = line.strip()
            if not in_field_table and not parsed_field_table and line.startswith("======"):
                in_field_table = True
                has_header = False
                continue
            if in_field_table and not has_header and line.startswith("======"):
                has_header = True
                continue
            if in_field_table and has_header and line.startswith("======"):
                in_field_table = False
                parsed_field_table = True
                continue
            if in_field_table and has_header:
                part = line.split(maxsplit=3)
                if len(part) < 3:
                    continue
                if not part[0].isnumeric():
                    continue
                offset = int(part[0])
                size = int(part[1]) if part[1].isnumeric() else 0
                dtype = part[2]
                label = part[3]

                code = label.split('[')[0].split('(')[0].split(',')[0].strip()
                code = code.lower().replace(" ", "_").replace('m/e', 'me').replace('?', '')
                ftype = "ftypes.STRING"
                valuestring = "nil"
                base = "base.NONE"
                if dtype == "u8":
                    ftype = "ftypes.UINT8"
                    base = "base.DEC"
                elif dtype == "u16":
                    ftype = "ftypes.UINT16"
                    base = "base.DEC"
                elif dtype == "u32":
                    ftype = "ftypes.UINT32"
                    base = "base.DEC"
                elif dtype == "i8":
                    ftype = "ftypes.INT8"
                    base = "base.DEC"
                elif dtype == "i16":
                    ftype = "ftypes.INT16"
                    base = "base.DEC"
                elif dtype == "i32":
                    ftype = "ftypes.INT32"
                    base = "base.DEC"
                elif dtype == "bool":
                    ftype = "base.UINT8"
                    base = "base.HEX"

                if 'mask' in label.lower() and base == "base.DEC":
                    base = "base.HEX"

                result += f'pf_fields["pf_cmd_{cmd}_{code}"] = ProtoField.new ("{label}", "atem.cmd.{cmd}.{code}", {ftype}, {valuestring}, {base})\n'
        result += "\n"
    return result


def make_cmd_labels():
    result = "local cmd_labels = {}\n"
    for label in pyatem.protocol.AtemProtocol.FIELDNAME_PRETTY:
        pretty = pyatem.protocol.AtemProtocol.FIELDNAME_PRETTY[label]
        result += f'cmd_labels["{label}"] = "{pretty}"\n'
    return result


def main():
    template_path = os.path.join(os.path.dirname(__file__), 'atem_dissector.tpl')
    with open(template_path) as handle:
        template = handle.read()

    fields = make_pf_fields()
    fields += make_cmd_labels()

    template = template.replace('@@FIELDS', fields)
    template = template.replace('@@REGS', '')
    template = template.replace('@@CMDS', '')
    print(template)


if __name__ == '__main__':
    main()



import argparse
import time

from pyatem.command import SendFairlightLevelsCommand, FairlightStripPropertiesCommand
from pyatem.field import FairlightMeterLevelsField
from pyatem.protocol import AtemProtocol

switcher: AtemProtocol = None

release_timer = time.time()


def on_disconnected():
    print("Hardware has disconnected")
    exit(0)


def run(device, main, sidechain, threshold, mode, release=0.3, ratio=2, duck_level=-30):
    global switcher

    # Parse the channel index, for split channels the format can be 1.1 for the right channel instead of just 1
    if '.' in main:
        main_index, main_sub = map(int, main.split('.'))
        main_strip = main
    else:
        main_index = int(main)
        main_sub = -1
        main_strip = f'{main}.0'

    if '.' in sidechain:
        side_index, side_sub = map(int, sidechain.split('.'))
    else:
        side_index = int(sidechain)
        side_sub = 0

    def on_connected():
        global switcher

        print("Connection successful")
        model = switcher.mixerstate['product-name']
        print(f"Detected hardware: {model.name}")
        fw = switcher.mixerstate['firmware-version']
        print(f"Firmware: {fw}")
        print()

        # Trigger receiving audio channels
        cmd = SendFairlightLevelsCommand(True)
        switcher.send_commands([cmd])

    def on_levels(level: FairlightMeterLevelsField):
        global release_timer
        if level.index == side_index and level.subchannel == side_sub:
            # Grab the output peak level
            val = sum(level.output[0:2]) / 2.0

            if val > threshold:
                gr = val - threshold
                if mode == 'compress':
                    new_vol = -int(gr * 100)
                else:
                    new_vol = duck_level

                if new_vol < switcher.mixerstate['fairlight-strip-properties'][main_strip].volume:
                    cmd = FairlightStripPropertiesCommand(main_index, main_sub, volume=-int(gr * 100))
                    switcher.send_commands([cmd])
                release_timer = time.time()
            elif switcher.mixerstate['fairlight-strip-properties'][main_strip].volume < 0 and (
                    time.time() - release_timer) > release:
                cur = switcher.mixerstate['fairlight-strip-properties'][main_strip].volume
                new_vol = int((cur * 0.9) * 1)
                cmd = FairlightStripPropertiesCommand(main_index, main_sub, volume=new_vol)
                switcher.send_commands([cmd])

    print(f"Connecting to {device}...")
    switcher = AtemProtocol(device)
    switcher.on('connected', on_connected)
    switcher.on('disconnected', on_disconnected)
    switcher.on('change:fairlight-meter-levels:*', on_levels)
    switcher.connect()
    while True:
        switcher.loop()


def main():
    parser = argparse.ArgumentParser(description="Duck channel volume on incoming audio of other channels")
    parser.add_argument('device', help="Device ip address or 'usb'")
    parser.add_argument('main', help='Channel number for the channel that needs to be reduced in level')
    parser.add_argument('sidechain', help='Channel number for the detection channel')
    parser.add_argument('--threshold', type=float, default=-30.0, help='Trigger level on the detection channel')
    parser.add_argument('--mode', choices=['compress', 'duck'], help='Method for selecting the reduction level',
                        default='duck')
    parser.add_argument('--level', type=float, default=-30, help='Target level for duck mode')
    parser.add_argument('--release', type=float, default=300, help='Release time in milliseconds')
    args = parser.parse_args()
    run(args.device, args.main, args.sidechain, args.threshold, mode=args.mode, duck_level=args.level,
        release=args.release / 1000)


if __name__ == '__main__':
    main()



import argparse
import signal

import tabulate

from pyatem.protocol import AtemProtocol
from pyatem.hexdump import hexdump

switcher = None
result = {}


def on_change(key, contents):
    global switcher
    global result
    if isinstance(contents, bytes):
        if key not in result:
            result[key] = set()
        result[key].add(contents)

    if key == 'multiviewer-input' or key == 'time':
        return
    
    print(key)
    if isinstance(contents, bytes):
        print('    ' + hexdump(contents, 'return').replace('\n', '\n    '))
    else:
        print('    ' + repr(contents))


def sigint_handler(signal, frame):
    print('--- [ Report ] ---')
    rows = []

    for field in result:
        d = ''
        for r in result[field]:
            d += repr(r) + '\n'
        rows.append((
            field,
            len(result[field]),
            d
        ))

    print(tabulate.tabulate(rows, headers=['field', 'count', 'data']))
    exit(0)


def on_connected():
    global switcher
    global testqueue
    global prepqueue
    global stats_start

    signal.signal(signal.SIGINT, sigint_handler)

    print("Connection successful")
    model = switcher.mixerstate['product-name']
    print(f"Detected hardware: {model.name}")


def on_disconnected():
    print("Hardware has disconnected")


def run(device):
    global switcher
    print(f"Connecting to {device}...")
    switcher = AtemProtocol(device)
    switcher.on('connected', on_connected)
    switcher.on('disconnected', on_disconnected)
    switcher.on('change', on_change)
    switcher.connect()
    while True:
        switcher.loop()


def main():
    parser = argparse.ArgumentParser(description="Find non-implemented fields in the connected hardware")
    parser.add_argument('device', help="Device ip address or 'usb'")
    args = parser.parse_args()
    run(args.device)


if __name__ == '__main__':
    main()



# Copyright 2021 - 2022, Martijn Braam and the OpenAtem contributors
# SPDX-License-Identifier: LGPL-3.0-only
import hashlib
import struct

from pyatem.media import rle_encode


class TransferTask:
    def __init__(self, store, slot, upload=False):
        self.tid = None
        self.state = None
        self.upload = upload

        self.store = store
        self.slot = slot

        self.data = None
        self.data_length = None
        self.hash = None

        self.send_length = None
        self.send_done = 0

        self.name = None
        self.description = None

    def calculate_hash(self):
        hasher = hashlib.md5(self.data)
        self.hash = hasher.digest()
        self.data_length = len(self.data)

    def compress(self):
        compressed = rle_encode(self.data)
        self.data = compressed
        self.send_length = len(self.data)

    def __repr__(self):
        direction = 'upload' if self.upload else 'download'
        return f'<TransferTask {direction} store={self.store} slot={self.slot}>'

    def to_tcp(self):
        name = self.name.encode() if self.name else b''
        description = self.description.encode() if self.description else b''
        header = struct.pack('>HH Hx? 64s 128s 16s II', self.tid or 0, self.store, self.slot, self.upload,
                             name, description, self.hash, self.send_length, self.data_length)

        # Large packets, let TCP fragmentation deal with it
        chunksize = 16000
        buffer = self.data
        packets = []
        while True:
            chunk = buffer[0:chunksize]
            buffer = buffer[chunksize:]
            packet = header + chunk
            packets.append((b'*XFR', packet))
            if len(buffer) == 0:
                break
        return packets

    @classmethod
    def from_tcp(cls, packet):
        header = struct.unpack_from('>HH Hx? 64s 128s 16s II', packet, 8)
        self = TransferTask(header[1], header[2])
        self.tid = header[0]
        self.upload = header[3]
        self.name = header[4].split(b'\x00')[0].decode()
        self.description = header[5].split(b'\x00')[0].decode()
        self.hash = header[6]
        self.send_length = header[7]
        self.data_length = header[8]
        self.data = packet[232:]
        return self


class TransferQueueFlushed:
    def __init__(self):
        pass



# Copyright 2021 - 2022, Martijn Braam and the OpenAtem contributors
# SPDX-License-Identifier: GPL-3.0-only
import gi
from gi.repository import GLib


class Service:
    def __init__(self, name, url):
        self.name = name
        self.url = url

    def variant(self):
        return GLib.Variant('a{sv}', {
            'name': GLib.Variant('s', self.name),
            'url': GLib.Variant('s', self.url),
        })


services = {
    "Youtube": [
        Service("Youtube Primary", "rtmp://a.rtmp.youtube.com/live2"),
        Service("Youtube Secondary", "rtmp://b.rtmp.youtube.com/live2?backup=1"),
    ],
    "Facebook": [
        Service("Facebook", "rtmps://live-api-s.facebook.com:443/rtmp"),
    ],
    "Twitch": [
        Service("Twitch auto location", "rtmp://live.twitch.tv/app"),
    ],
    "Twitter / Periscope": [
        Service("Periscope Tokyo", "rtmp://jp.pscp.tv:80/x"),
        Service("Periscope Seoul", "rtmp://kr.pscp.tv:80/x"),
        Service("Periscope Mumbai", "rtmp://in.pscp.tv:80/x"),
        Service("Periscope Singapore", "rtmp://sg.pscp.tv:80/x"),
        Service("Periscope Sydney", "rtmp://au.pscp.tv:80/x"),
        Service("Periscope Canada", "rtmp://va.pscp.tv:80/x"),
        Service("Periscope Frankfurt", "rtmp://de.pscp.tv:80/x"),
        Service("Periscope Ireland", "rtmp://ie.pscp.tv:80/x"),
        Service("Periscope São Paulo", "rtmp://br.pscp.tv:80/x"),
        Service("Periscope N. Virginia", "rtmp://va.pscp.tv:80/x"),
        Service("Periscope Ohio", "rtmp://va.pscp.tv:80/x"),
        Service("Periscope Oregon", "rtmp://or.pscp.tv:80/x"),
        Service("Periscope N. California", "rtmp://ca.pscp.tv:80/x"),
    ],
    "Restream.io": [
        Service("Restream.io auto location", "rtmp://live.restream.io/live")
    ],
}



import threading
import logging
from functools import partial
from time import sleep

from .error import DependencyError
import pyatem.command as commandmodule

try:
    import rtmidi
    from rtmidi.midiutil import open_midiinput, open_midioutput, list_available_ports
    from rtmidi.midiconstants import NOTE_ON, NOTE_OFF, PITCH_BEND
except ModuleNotFoundError:
    rtmidi = None


class MidiFrontendThread(threading.Thread):
    def __init__(self, config, threadlist):
        threading.Thread.__init__(self)
        if rtmidi is None:
            raise DependencyError("The rtmidi library is not available")
        self.name = 'midi.' + str(config['bind'])
        self.bind = str(config['bind'])
        self.config = config
        self.threadlist = threadlist
        self.status = 'initializing...'
        self.error = None

        self.port = None
        self.output = None
        self.input = None
        self.map = {}
        self.reverse = {}
        self.dump = config['dump'] if 'dump' in config else False

        self.eventname = {
            'CC': 11,
            'NOTE-ON': 9,
            'NOTE-OFF': 8,
            'PITCH-BEND': 14,
        }

        # Parse the configured midi events and create the midi->switcher mapping
        for ckey in config:
            if '/' not in ckey:
                continue
            part = ckey.split('/')
            if part[0] == '*':
                channel = None
            else:
                channel = int(part[0])
            event = part[1]
            if event.isnumeric():
                event = int(part[1])
            else:
                event = self.eventname[event]
            key = None
            if len(part) > 2:
                key = int(part[2])
            if len(part) > 3:
                value = int(part[3])
            else:
                value = None
            mkey = (channel, event, key, value)

            fieldname = config[ckey]['field']
            classname = fieldname.title().replace('-', '') + "Command"
            if not hasattr(commandmodule, classname):
                logging.error(f'unrecognized command {fieldname}')
                self.status = 'error'
                self.error = 'config error'
                return
            arguments = {}
            for argname in config[ckey]:
                if argname in ['hardware', 'field']:
                    continue
                arguments[argname] = config[ckey][argname]
            cmd = getattr(commandmodule, classname)(**arguments)
            action = (config[ckey]['hardware'], cmd)

            if mkey not in self.map:
                self.map[mkey] = []
            self.map[mkey].append(action)

        # Parse the switcher events and create the switcher->midi mapping
        for ckey in config:
            if ':' not in ckey:
                continue

            name, field, raw = ckey.split(':', maxsplit=2)
            filters = {}
            raw = raw.split(':')
            for i, f in enumerate(raw):
                if '=' in f:
                    k, v = f.split('=', maxsplit=1)
                    v = int(v)
                else:
                    k = f
                    v = None
                filters[k] = v
                if i == len(raw) - 1:
                    filters['_val'] = k

            if name not in self.reverse:
                self.reverse[name] = {}

            if field not in self.reverse[name]:
                self.reverse[name][field] = []
            filters['_action'] = config[ckey]
            self.reverse[name][field].append(filters)

        # Subscribe to all requested switchers
        hw_names = set()
        for ckey in config:
            if ':' not in ckey:
                continue
            name, _ = ckey.split(':', maxsplit=1)
            hw_names.add(name)
        sleep(0.1)
        for hw in hw_names:
            sw = self.threadlist['hardware'][hw].switcher
            sw.on('change', partial(self.on_switcher_changed, hw))
            sw.on('connected', partial(self.on_switcher_connected, hw))

    def run(self):
        logging.info('MIDI frontend run')

        temp = rtmidi.MidiIn()
        ports = temp.get_ports()
        options = []
        for port in ports:
            if 'Midi Through' in port:
                continue
            options.append(port)
            if self.bind == 'any':
                break
            if self.bind == port:
                break
        else:
            if self.bind == 'any':
                logging.error(f'Could not find any midi devices to bind to')
            else:
                logging.error(f'Could not bind to midi device "{self.bind}"')
                devices = '", "'.join(options)
                logging.error(f'Midi devices present: ["{devices}"]')
            self.status = "error"
            self.error = "hardware not present"
            return

        self.port = port
        midiin, port_name = open_midiinput(self.port, client_name="OpenSwitcher Proxy")
        midiout, port_name = open_midioutput(self.port, client_name="OpenSwitcher Proxy")

        self.output = midiout
        self.input = midiin
        midiin.set_callback(self.on_midi_in)

        hw_names = set()
        for ckey in self.config:
            if ':' not in ckey:
                continue
            name, _ = ckey.split(':', maxsplit=1)
            hw_names.add(name)

        for hw in hw_names:
            sw = self.threadlist['hardware'][hw].switcher
            sw.on('change', partial(self.on_switcher_changed, hw))
            sw.on('connected', partial(self.on_switcher_connected, hw))

            if self.threadlist['hardware'][hw].status == 'connected':
                # Hardware is already connected at this point, re-generate the initial data
                self.on_switcher_connected(hw)

        self.status = 'running'

    def get_events(self, channel, event, key, value):
        result = []
        mkey = (channel, event, key, value)
        if mkey in self.map:
            result.extend(self.map[mkey])
        mkey = (channel, event, key, None)
        if mkey in self.map:
            result.extend(self.map[mkey])
        mkey = (None, event, key, value)
        if mkey in self.map:
            result.extend(self.map[mkey])
        mkey = (None, event, key, None)
        if mkey in self.map:
            result.extend(self.map[mkey])
        return result

    def on_midi_in(self, raw, data=None):
        event_raw, key, value = raw[0]
        event = event_raw >> 4
        channel = event_raw & 0b00001111

        if event == 14:
            value = (value << 7) | key
            key = None

        if self.dump:
            ename = f'EVENT-{event}'
            for e in self.eventname:
                if event == self.eventname[e]:
                    ename = e
            if ename == 'PITCH-BEND':
                print(f'{channel}/{ename} = {value}')
            else:
                print(f'{channel}/{ename}/{key} = {value}')
        for action in self.get_events(channel, event, key, value):
            command = action[1]
            if event == 14:
                command.position = int(value / (2 ** 14) * 10000)
            self.threadlist['hardware'][action[0]].switcher.send_commands([command])

    def on_switcher_changed(self, hw, field, value):
        if hw not in self.reverse:
            return
        if field not in self.reverse[hw]:
            return
        for f in self.reverse[hw][field]:
            match = True
            state = False
            for filter_field in f:
                if filter_field in ["_action", "_val"]:
                    continue
                filter_value = f[filter_field]
                real_value = getattr(value, filter_field)

                if filter_field == f['_val'] and filter_value is not None:
                    state = real_value == filter_value
                    continue
                elif filter_field == f['_val'] and filter_value is None:
                    state = real_value
                elif real_value != filter_value:
                    match = False
                    break

            if match:
                action = f['_action']

                if action['event'] == 'note-on':
                    channel = (action['channel'] if 'channel' in action else 1) - 1
                    message = [
                        NOTE_ON | channel,
                        action['key'],
                        action['on'] if state else action['off']
                    ]
                elif action['event'] == 'pitch-bend':
                    channel = (action['channel'] if 'channel' in action else 1) - 1
                    midivalue = (state - action['min']) / (action['max'] - action['min'])
                    midivalue = int(midivalue * (2 ** 14))
                    lsb = midivalue & 0b01111111
                    msb = (midivalue >> 7) & 0b01111111
                    message = [
                        PITCH_BEND | channel,
                        lsb,
                        msb
                    ]
                else:
                    raise ValueError(f"Unknown action '{action['event']}'")
                self.output.send_message(message)

    def on_switcher_connected(self, hw):
        sw = self.threadlist['hardware'][hw].switcher
        items = list(sw.mixerstate.items())
        for field, value in items:
            if isinstance(value, dict):
                for k in value:
                    v = value[k]
                    self.on_switcher_changed(hw, field, v)
            else:
                self.on_switcher_changed(hw, field, value)

    def get_status(self):
        if self.status == 'error':
            return f'{self.status}, {self.error}'
        else:
            return f'{self.status} ({self.port})'



# Copyright 2021 - 2022, Martijn Braam and the OpenAtem contributors
# SPDX-License-Identifier: GPL-3.0-only
import gi

from pyatem.command import MultiviewInputCommand
from pyatem.field import InputPropertiesField
from pyatem.macro import decode_macro, encode_macro, encode_macroscript

gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, GLib, GObject, Gio, Gdk

gi.require_version('Handy', '1')
from gi.repository import Handy


class MacroEditorWindow:
    def __init__(self, parent, application, connection, index, raw):
        self.application = application
        self.connection = connection

        self.index = index
        self.raw = raw

        builder = Gtk.Builder()
        builder.set_translation_domain("openswitcher")
        builder.add_from_resource('/nl/brixit/switcher/ui/macro-editor.glade')
        builder.connect_signals(self)
        css = Gio.resources_lookup_data("/nl/brixit/switcher/ui/style.css", 0)

        self.provider = Gtk.CssProvider()
        self.provider.load_from_data(css.get_data())

        self.window = builder.get_object("window")
        self.headerbar = builder.get_object("headerbar")
        self.main_stack = builder.get_object("main_stack")
        self.sourcecode = builder.get_object("sourcecode")
        self.sourcebuffer = builder.get_object("sourcebuffer")
        self.window.set_application(self.application)

        self.actions = builder.get_object("actions")

        self.apply_css(self.window, self.provider)

        self.window.set_transient_for(parent)
        self.window.set_modal(True)

        macro = self.connection.mixer.mixerstate['macro-properties'][index]
        self.headerbar.set_subtitle(macro.name.decode())

        ma = decode_macro(raw)
        source = encode_macroscript(ma)
        self.sourcebuffer.set_text(source)
        for action in ma:
            frame = Gtk.Frame()
            frame.get_style_context().add_class('view')
            grid = Gtk.Grid()
            grid.set_margin_top(8)
            grid.set_margin_bottom(8)
            grid.set_margin_start(8)
            grid.set_margin_end(8)
            grid.set_column_spacing(8)
            grid.set_row_spacing(8)
            frame.add(grid)
            name = Gtk.Label(action.__class__.NAME)
            name.set_xalign(0.0)
            grid.attach(name, 0, 0, 2, 1)

            top = 1
            for a in action.widgets:
                for attribute, datatype, label, properties in action.widgets[a]:
                    field_label = Gtk.Label(label)
                    field_label.get_style_context().add_class('dim-label')
                    field_label.set_xalign(1.0)
                    grid.attach(field_label, 0, top, 1, 1)

                    widget = None
                    if datatype == 'framecount':
                        widget = Gtk.SpinButton()
                        adjustment = Gtk.Adjustment(getattr(action, attribute), 0, 250, 1, 30, 1)
                        widget.adjustment = adjustment
                        widget.set_adjustment(adjustment)
                    elif datatype == 'number':
                        widget = Gtk.SpinButton()

                        value = getattr(action, attribute)
                        if 'offset' in properties:
                            value += properties['offset']

                        adjustment = Gtk.Adjustment(value, properties['min'], properties['max'], 1, 10, 1)
                        widget.adjustment = adjustment
                        widget.set_adjustment(adjustment)
                    elif datatype == 'source':
                        widget = Gtk.ComboBox()

                    if widget:
                        grid.attach(widget, 1, top, 1, 1)

                    top += 1

            self.actions.add(frame)

        self.window.show_all()

    def apply_css(self, widget, provider):
        Gtk.StyleContext.add_provider(widget.get_style_context(),
                                      provider,
                                      Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION)

        if isinstance(widget, Gtk.Container):
            widget.forall(self.apply_css, provider)

    def on_cancel_clicked(self, *args):
        self.window.close()

    def on_save_clicked(self, *args):
        self.window.close()

    def on_source_toggled(self, widget, *args):
        if widget.get_active():
            self.main_stack.set_visible_child_name("code")
        else:
            self.main_stack.set_visible_child_name("blocks")



# Copyright 2021 - 2022, Martijn Braam and the OpenAtem contributors
# SPDX-License-Identifier: GPL-3.0-only
import argparse
import os
import gi

from bmd_setup.window import SetupWindow

gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, Gio, GLib

gi.require_version('Handy', '1')
from gi.repository import Handy


class SetupApplication(Gtk.Application):
    def __init__(self, application_id, flags):
        Gtk.Application.__init__(self, application_id=application_id, flags=flags)
        GLib.set_prgname(application_id)
        self.connect("activate", self.new_window)

    def new_window(self, *args):
        SetupWindow(self)


def main(version):
    Handy.init()
    app = SetupApplication("nl.brixit.Setup", Gio.ApplicationFlags.FLAGS_NONE)
    app.run()


if __name__ == '__main__':
    main('')



# Copyright 2021 - 2022, Martijn Braam and the OpenAtem contributors
# SPDX-License-Identifier: LGPL-3.0-only
import math
import sys
import inspect

from pyatem.command import CameraControlCommand

_cache = {}

VOID = 0
BOOL = 0
INT8 = 1
INT16 = 2
INT32 = 3
INT64 = 4
UTF8 = 5
FIXED16 = 128


class CameraControlData:
    CATEGORY = -1
    PARAMETER = -1
    DATATYPE = -1
    DESCRIPTIONS = [""]
    KEYS = [""]

    def __init__(self, destination=None, **kwargs):
        self.data = None
        self.destination = destination

        if len(kwargs) > 0:
            self.data = [0] * len(self.DESCRIPTIONS)
            for key in kwargs:
                offset = self.KEYS.index(key)
                self.data[offset] = kwargs[key]

    @classmethod
    def from_data(cls, data):
        """
        :type data: pyatem.field.CameraControlDataPacketField
        """
        global _cache
        if len(_cache) == 0:
            current_module = sys.modules[__name__]
            for name, dcls in inspect.getmembers(current_module):
                if hasattr(dcls, 'CATEGORY') and dcls.CATEGORY > -1:
                    _cache[(dcls.CATEGORY, dcls.PARAMETER)] = dcls

        if (data.category, data.parameter) in _cache:
            instance = _cache[(data.category, data.parameter)]()
            instance.data = data.data
            instance.destination = data.destination
            instance.decode()
            return instance
        return None

    def to_command(self, relative=False):
        cmd = CameraControlCommand(self.destination, self.CATEGORY, self.PARAMETER, relative, self.DATATYPE, self.data)
        return cmd

    def decode(self):
        pass

    def __repr__(self):
        values = ""
        if self.data is not None:
            for i, value in enumerate(self.data):
                values += f' {self.KEYS[i]}={value}{self.DESCRIPTIONS[i]}'
        return f'<{self.__class__.__name__} dest={self.destination} id={self.CATEGORY}.{self.PARAMETER}{values}>'


class Focus(CameraControlData):
    CATEGORY = 0
    PARAMETER = 0
    DATATYPE = FIXED16
    KEYS = ["distance"]


class TriggerAutofocus(CameraControlData):
    CATEGORY = 0
    PARAMETER = 1
    DATATYPE = VOID
    DESCRIPTIONS = []


class ApertureFStop(CameraControlData):
    CATEGORY = 0
    PARAMETER = 2
    DATATYPE = FIXED16
    KEYS = ["fnumber"]

    def decode(self):
        self.data[0] = math.sqrt(2 ** self.data[0])


class ApertureNormalized(CameraControlData):
    CATEGORY = 0
    PARAMETER = 3
    DATATYPE = FIXED16
    KEYS = ["aperture"]


class ApertureOrdinal(CameraControlData):
    CATEGORY = 0
    PARAMETER = 3
    DATATYPE = FIXED16
    KEYS = ["aperture"]


class TriggerAutoaperture(CameraControlData):
    CATEGORY = 0
    PARAMETER = 5
    DATATYPE = VOID
    DESCRIPTIONS = []


class OIS(CameraControlData):
    CATEGORY = 0
    PARAMETER = 6
    DATATYPE = BOOL
    KEYS = ["enabled"]


class AbsoluteZoom(CameraControlData):
    CATEGORY = 0
    PARAMETER = 7
    DATATYPE = INT16
    KEYS = ["zoom"]
    DESCRIPTIONS = ["mm"]


class AbsoluteZoomNormalized(CameraControlData):
    CATEGORY = 0
    PARAMETER = 8
    DATATYPE = FIXED16
    KEYS = ["zoom"]


class ContinuousZoom(CameraControlData):
    CATEGORY = 0
    PARAMETER = 9
    DATATYPE = FIXED16
    KEYS = ["rate"]


class VideoMode(CameraControlData):
    CATEGORY = 1
    PARAMETER = 0
    DATATYPE = INT8
    KEYS = ["framerate", "mrate", "dimensions", "interlaced", "colorspace"]
    DESCRIPTIONS = ["fps", "", "", "", ""]


class Gain(CameraControlData):
    CATEGORY = 1
    PARAMETER = 1
    DATATYPE = INT8
    KEYS = ["ISO"]

    def decode(self):
        self.data = [self.data[0] * 100]


class WhiteBalance(CameraControlData):
    CATEGORY = 1
    PARAMETER = 2
    DATATYPE = INT16
    KEYS = ["temperature", "tint"]
    DESCRIPTIONS = ["k", ""]


class TriggerAutowhitebalance(CameraControlData):
    CATEGORY = 1
    PARAMETER = 3
    DATATYPE = VOID
    DESCRIPTIONS = []


class TriggerRestorewhitebalance(CameraControlData):
    CATEGORY = 1
    PARAMETER = 4
    DATATYPE = VOID
    DESCRIPTIONS = []


class Exposure(CameraControlData):
    CATEGORY = 1
    PARAMETER = 5
    DATATYPE = INT32
    KEYS = ["time"]
    DESCRIPTIONS = ["us"]


class DynamicRangeMode(CameraControlData):
    CATEGORY = 1
    PARAMETER = 7
    DATATYPE = INT8
    KEYS = ["mode"]


class VideoSharpening(CameraControlData):
    CATEGORY = 1
    PARAMETER = 8
    DATATYPE = INT8
    KEYS = ["level"]


class RecordingFormat(CameraControlData):
    CATEGORY = 1
    PARAMETER = 9
    DATATYPE = INT16
    KEYS = ["file-fps", "sensor-fps", "width", "height", "flags"]
    DESCRIPTIONS = ["", "", "", "", ""]


class AutoExposureMode(CameraControlData):
    CATEGORY = 1
    PARAMETER = 10
    DATATYPE = INT8
    KEYS = ["mode"]


class ShutterAngle(CameraControlData):
    CATEGORY = 1
    PARAMETER = 11
    DATATYPE = INT32
    KEYS = ["angle"]
    DESCRIPTIONS = ["deg"]

    def decode(self):
        self.data = [self.data[0] / 100]


class ShutterSpeed(CameraControlData):
    CATEGORY = 1
    PARAMETER = 12
    DATATYPE = INT32
    KEYS = ["speed"]


class GainDB(CameraControlData):
    CATEGORY = 1
    PARAMETER = 13
    DATATYPE = INT8
    KEYS = ["gain"]
    DESCRIPTIONS = ["dB"]


class ISO(CameraControlData):
    CATEGORY = 1
    PARAMETER = 14
    DATATYPE = INT32
    KEYS = ["iso"]
    DESCRIPTIONS = ["ISO"]


class MicLevel(CameraControlData):
    CATEGORY = 2
    PARAMETER = 0
    DATATYPE = FIXED16
    KEYS = ["level"]


class HeadphoneLevel(CameraControlData):
    CATEGORY = 2
    PARAMETER = 1
    DATATYPE = FIXED16
    KEYS = ["level"]


class HeadphoneProgramMix(CameraControlData):
    CATEGORY = 2
    PARAMETER = 2
    DATATYPE = FIXED16
    KEYS = ["mix"]


class SpeakerLevel(CameraControlData):
    CATEGORY = 2
    PARAMETER = 3
    DATATYPE = FIXED16
    KEYS = ["level"]


class AudioInputType(CameraControlData):
    CATEGORY = 2
    PARAMETER = 4
    DATATYPE = INT8
    KEYS = ["type"]


class AudioInputLevels(CameraControlData):
    CATEGORY = 2
    PARAMETER = 5
    DATATYPE = FIXED16
    KEYS = ["left", "right"]


class PhantomPower(CameraControlData):
    CATEGORY = 2
    PARAMETER = 6
    DATATYPE = BOOL
    KEYS = ["enabled"]


class OutputOverlay(CameraControlData):
    CATEGORY = 3
    PARAMETER = 0
    DATATYPE = INT16
    KEYS = "flag"


class OutputFrameGuideStyle(CameraControlData):
    CATEGORY = 3
    PARAMETER = 1
    DATATYPE = INT8
    KEYS = "style"


class OutputFrameGuideOpacity(CameraControlData):
    CATEGORY = 3
    PARAMETER = 2
    DATATYPE = FIXED16
    KEYS = "opacity"


class OutputOverlays(CameraControlData):
    CATEGORY = 3
    PARAMETER = 3
    DATATYPE = INT8
    KEYS = ["style", "opacity", "safearea", "gridstyle"]
    DESCRIPTIONS = ["", "%", "%", ""]


class DisplayBrightness(CameraControlData):
    CATEGORY = 4
    PARAMETER = 0
    DATATYPE = FIXED16
    KEYS = ["brightness"]


class DisplayOverlay(CameraControlData):
    CATEGORY = 4
    PARAMETER = 1
    DATATYPE = INT16
    KEYS = ["bitfield"]


class DisplayZebraLevel(CameraControlData):
    CATEGORY = 4
    PARAMETER = 2
    DATATYPE = FIXED16
    KEYS = ["level"]


class DisplayPeakingLevel(CameraControlData):
    CATEGORY = 4
    PARAMETER = 3
    DATATYPE = FIXED16
    KEYS = ["level"]


class DisplayColorBarsTime(CameraControlData):
    CATEGORY = 4
    PARAMETER = 4
    DATATYPE = INT8
    KEYS = ["seconds"]


class DisplayFocusAssist(CameraControlData):
    CATEGORY = 4
    PARAMETER = 5
    DATATYPE = INT8
    KEYS = ["method", "color"]
    DESCRIPTIONS = ["", ""]


class TallyBrightness(CameraControlData):
    CATEGORY = 5
    PARAMETER = 0
    DATATYPE = FIXED16
    KEYS = ["brightness"]


class TallyFrontBrightness(CameraControlData):
    CATEGORY = 5
    PARAMETER = 1
    DATATYPE = FIXED16
    KEYS = ["brightness"]


class TallyRearBrightness(CameraControlData):
    CATEGORY = 5
    PARAMETER = 2
    DATATYPE = FIXED16
    KEYS = ["brightness"]


class ReferenceSource(CameraControlData):
    CATEGORY = 6
    PARAMETER = 0
    DATATYPE = INT8
    KEYS = ["source"]


class ReferenceOffset(CameraControlData):
    CATEGORY = 6
    PARAMETER = 1
    DATATYPE = INT32
    KEYS = ["offset"]
    DESCRIPTIONS = ["px"]


class RealtimeClock(CameraControlData):
    CATEGORY = 7
    PARAMETER = 0
    DATATYPE = INT32
    KEYS = ["time", "date"]
    DESCRIPTIONS = ["", " as BCD"]


class SystemLanguage(CameraControlData):
    CATEGORY = 7
    PARAMETER = 1
    DATATYPE = UTF8
    KEYS = ["lang"]


class Timezone(CameraControlData):
    CATEGORY = 7
    PARAMETER = 2
    DATATYPE = INT32
    KEYS = ["offset"]
    DESCRIPTIONS = [" minutes"]


class Location(CameraControlData):
    CATEGORY = 7
    PARAMETER = 3
    DATATYPE = INT64
    KEYS = ["latitude", "longitude"]
    DESCRIPTIONS = ["", " as BCD"]


class LiftAdjust(CameraControlData):
    CATEGORY = 8
    PARAMETER = 0
    DATATYPE = FIXED16
    KEYS = ["red", "green", "blue", "luma"]
    DESCRIPTIONS = ["", "", "", ""]


class GammaAdjust(CameraControlData):
    CATEGORY = 8
    PARAMETER = 1
    DATATYPE = FIXED16
    KEYS = ["red", "green", "blue", "luma"]
    DESCRIPTIONS = ["", "", "", ""]


class GainAdjust(CameraControlData):
    CATEGORY = 8
    PARAMETER = 2
    DATATYPE = FIXED16
    KEYS = ["red", "green", "blue", "luma"]
    DESCRIPTIONS = ["", "", "", ""]


class OffsetAdjust(CameraControlData):
    CATEGORY = 8
    PARAMETER = 3
    DATATYPE = FIXED16
    KEYS = ["red", "green", "blue", "luma"]
    DESCRIPTIONS = ["", "", "", ""]


class ContrastAdjust(CameraControlData):
    CATEGORY = 8
    PARAMETER = 4
    DATATYPE = FIXED16
    KEYS = ["pivot", "adjust"]
    DESCRIPTIONS = ["", ""]


class LumaMix(CameraControlData):
    CATEGORY = 8
    PARAMETER = 5
    DATATYPE = FIXED16
    KEYS = ["mix"]


class ColorAdjust(CameraControlData):
    CATEGORY = 8
    PARAMETER = 6
    DATATYPE = FIXED16
    KEYS = ["hue", "saturation"]
    DESCRIPTIONS = ["", ""]


class TriggerColorReset(CameraControlData):
    CATEGORY = 8
    PARAMETER = 7
    DATATYPE = VOID
    DESCRIPTIONS = []


class Codec(CameraControlData):
    CATEGORY = 10
    PARAMETER = 0
    DATATYPE = INT8
    KEYS = ["codec", "variant"]
    DESCRIPTIONS = ["", ""]


class TransportMode(CameraControlData):
    CATEGORY = 10
    PARAMETER = 1
    DATATYPE = INT8
    KEYS = ["mode", "speed", "flags", "storage"]
    DESCRIPTIONS = ["", "x", "", ""]


class PanTiltVelocity(CameraControlData):
    CATEGORY = 11
    PARAMETER = 0
    DATATYPE = FIXED16
    KEYS = ["pan", "tilt"]
    DESCRIPTIONS = ["", ""]


class PositionPreset(CameraControlData):
    CATEGORY = 11
    PARAMETER = 1
    DATATYPE = INT8
    KEYS = ["command", "slot"]
    DESCRIPTIONS = ["", ""]



# Copyright 2022 - 2022, Martijn Braam and the OpenAtem contributors
# SPDX-License-Identifier: LGPL-3.0-only
import os.path

from pyatem.converters.lut import load_cube, lut_to_bmd17

fixtures_dir = os.environ.get('TEST_FIXTURES', os.path.join(os.path.dirname(__file__), '../fixtures'))


def test_load_cube():
    result = load_cube(os.path.join(fixtures_dir, 'linear-17.cube'))
    assert result.title == "Generated by Resolve"
    assert result.keywords['LUT_3D_SIZE'] == 17
    assert len(result.table) == 4913


def test_generate_bmd17():
    lut = load_cube(os.path.join(fixtures_dir, 'linear-17.cube'))
    result = lut_to_bmd17(lut)

    with open(os.path.join(fixtures_dir, 'linear-17.bin'), 'rb') as handle:
        reference = handle.read()

    assert result == reference



# Copyright 2022 - 2022, Martijn Braam and the OpenAtem contributors
# SPDX-License-Identifier: LGPL-3.0-only
import argparse
import math
import os.path

import tqdm

import pyatem.converters.converter as conv
from pyatem.converters.protocol import Converter, WValueProtoConverter, Field

annotate = False


def enumerate_hardware():
    classes = []
    for name, cls in conv.__dict__.items():
        if isinstance(cls, type) and name not in ['Field', 'ValueField', 'Converter']:
            classes.append(cls)
    return classes


def enumerate_serials():
    for model in enumerate_hardware():
        pass


def dump_memory(device, target):
    if not isinstance(device, WValueProtoConverter):
        print("Operation not supported for this hardware")
        return

    memory = b''
    for offset in tqdm.tqdm(range(0x0000, 0xFFFF, 4)):
        chunk = bytes(device.handle.ctrl_transfer(bmRequestType=0xc1,
                                                  bRequest=83,
                                                  wValue=offset,
                                                  wIndex=0,
                                                  data_or_wLength=4))
        memory += chunk

    with open(target, 'wb') as handle:
        handle.write(memory)


def print_option(field, value):
    assert (isinstance(field, Field))
    if annotate and field.code is not None:
        print(f"{field.label}:  {value} (--{field.code})")
    else:
        print(f"{field.label}:  {value}")


def main():
    parser = argparse.ArgumentParser(description="Blackmagic Design Converter Setup")
    generic = parser.add_argument_group("Generic arguments")
    generic.add_argument('--factory-reset', action='store_true', help='Run the factory reset operation', dest='reset')
    generic.add_argument('--dump', help='Dump internal memory to file')
    generic.add_argument('--annotate', help='Show argument names', action='store_true')
    generic.add_argument('--serial', help='Select one specific converter')

    opt = parser.add_argument_group('Write converter setting')

    codes = set()
    codelut = {}
    for device in enumerate_hardware():
        for arg in device.FIELDS:
            if arg.code is not None:
                codes.add(arg.code)
                codelut[arg.code] = arg

    for item in sorted(codes):
        arg_type = None
        arg_choices = None
        field = codelut[item]
        if isinstance(field.mapping, dict):
            arg_type = str
            arg_choices = []
            for option in field.mapping.values():
                arg_choices.append(option[0])
        elif field.mapping == 'dB':
            arg_type = float
        opt.add_argument('--' + item, type=arg_type, choices=arg_choices)

    args = parser.parse_args()

    global annotate
    if args.annotate:
        annotate = True

    existing = []
    for device in enumerate_hardware():
        for serial in device.enumerate(device):
            existing.append((device, serial))

    if len(existing) == 0:
        print("No supported hardware detected")
        exit(1)
    elif len(existing) > 1:
        if args.serial is None:
            print("Multiple converters found")
            for dev, serial in existing:
                print(f" - [{serial}] {dev.NAME}")
            print("Select one by specifying the --serial argument")
            exit(1)
        for dev, serial in existing:
            if serial == args.serial:
                deviceclass = dev
                break
        else:
            print("Device not found")
            exit(1)
    else:
        deviceclass = existing[0][0]

    possible_codes = set()
    for item in deviceclass.FIELDS:
        if item.code is not None:
            possible_codes.add(item.code)

    print(f"Product:  {deviceclass.NAME}")

    to_write = []
    for code in codes:
        key = code.replace('-', '_')
        if getattr(args, key) is not None:
            if code not in possible_codes:
                print(f"Option --{code} is not valid for this device")
                exit(1)
            for item in deviceclass.FIELDS:
                if item.code == code:
                    to_write.append((item, getattr(args, key)))

    device = deviceclass()
    device.connect(serial=args.serial)
    if device.get_status() == Converter.STATUS_NEED_POWER:
        print("Converter needs power plugged in to be configured")
        exit(1)

    if args.dump:
        dump_memory(device, args.dump)
        exit(0)

    if len(to_write) > 0:
        print('===[ Write values ]===')
    lut_writes = []
    for field, value in to_write:
        if field.dtype == open:
            lut_writes.append((field, value))
            continue
        elif isinstance(field.mapping, dict):
            for raw_val in field.mapping:
                if field.mapping[raw_val][0] == value:
                    value = raw_val
                    break

        print_option(field, value)
        device.set_value(field, value)

    last_section = None
    device_state = device.get_state()
    for field_config in device.FIELDS:
        field = device_state[field_config.key]
        if field is None:
            continue
        if field.section != last_section:
            print(f"\n===[ {field.section} ]===")
            last_section = field.section
        value = field.value
        if field.dtype == open:
            continue
        if field.mapping is None:
            print_option(field, value)
        elif isinstance(field.mapping, dict):
            print(field.label + ':')
            got_check = False
            for key, display in field.mapping.items():
                x = 'x' if key == value else ' '
                if key == value:
                    got_check = True
                if annotate:
                    print(f'    [{x}] {display[1]} (--{field.code} {display[0]})')
                else:
                    print(f'    [{x}] {display[1]}')
            if not got_check:
                print(f"    Unknown value: {value}")
        elif isinstance(field.mapping, str):
            if field.mapping == 'dB':
                if value > float('-inf'):
                    print_option(field, f' {value:.2f} dB')
                else:
                    print_option(field, 'off')

    for field, value in lut_writes:
        title = os.path.basename(value)
        print(f"Uploading LUT '{title}'...")
        device.set_lut(field.key, value)
        print("done")

    if args.reset:
        print()
        print("Factory reset requested, press enter to reset all settings for this device or press ctrl+c to cancel")
        input()
        print("Executing factory reset")
        device.factory_reset()


if __name__ == '__main__':
    main()



# Copyright 2021 - 2022, Martijn Braam and the OpenAtem contributors
# SPDX-License-Identifier: GPL-3.0-only
import logging
import platform
from gi.repository import Gtk, GObject, Gio, GLib
import xml.etree.ElementTree as ET
import os

_PERFORM_MANUAL_TRANSLATION = platform.system() == "Windows"

class TemplateLocale(Gtk.Template):

    def __call__(self, cls):
        if not _PERFORM_MANUAL_TRANSLATION:
            logging.getLogger('TemplateLocale').debug("Localize is not needed")

            return super(TemplateLocale, self).__call__(cls)

        element = Gio.resources_lookup_data(self.resource_path, Gio.ResourceLookupFlags.NONE).get_data().decode('utf-8')
        tree = ET.fromstring(element)
        for node in tree.iter():
            context = ''
            if 'context' in node.attrib:
                context = node.attrib['context'] + "\x04"
            if 'translatable' in node.attrib:
                node.text = _(context + node.text)
        self.string = ET.tostring(tree, encoding='unicode', method='xml')
        self.resource_path = None

        return Gtk.Template.__call__(self, cls)



# Copyright 2022 - 2022, Martijn Braam and the OpenAtem contributors
# SPDX-License-Identifier: LGPL-3.0-only
from setuptools import setup, Extension

setup(
    name='pyatem',
    version='0.13.0',
    packages=['pyatem', 'pyatem.converters'],
    ext_modules=[Extension('pyatem.mediaconvert', ['pyatem/mediaconvertmodule.c'])],
    url='https://git.sr.ht/~martijnbraam/pyatem',
    license='LGPL3',
    author='Martijn Braam',
    author_email='martijn@brixit.nl',
    description='Library implementing the Blackmagic Design Atem switcher protocol',
    long_description=open("README.md").read(),
    classifiers=[
        'License :: OSI Approved :: GNU Lesser General Public License v3 (LGPLv3)',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Operating System :: POSIX :: Linux',
    ],
)



# Copyright 2025, Martijn Braam and the OpenAtem contributors
# SPDX-License-Identifier: GPL-3.0-only
import glob
import json
import os
import uuid

from gi.repository import Gtk, GObject, Gdk, GLib, Gio
import pyatem.field as fieldmodule
from gtk_switcher.adjustmententry import AdjustmentEntry
from gtk_switcher.presetwindow import PresetWindow

from pyatem.protocol import AtemProtocol

_palette_field = {}


class EventGate:
    def __init__(self, reference):
        self.ref = reference

    def __enter__(self):
        self.ref.model_changing = True

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.ref.model_changing = False


class RadioProxy:
    def __init__(self, parent, radios):
        self.parent = parent
        self.radios = {}
        for r in radios:
            self.radios[r.value] = r

    def set_value(self, value):
        with self.parent.model:
            self.radios[value].set_active(True)


class PaletteGroup:
    def __init__(self, panel):
        self.panel = panel
        self.stack: Gtk.Stack = None
        self.grid = None
        self.name = None
        self.title = None
        self.controls = []
        self._row = 0

    def add_page(self, name, title):
        if self.stack is None:
            raise ValueError("This is not a toplevel group")
        grid = Gtk.Grid()
        grid.set_column_spacing(12)
        grid.set_row_spacing(12)
        grid.set_margin_top(12)
        grid.set_margin_bottom(12)
        grid.set_margin_start(12)
        grid.set_margin_end(12)

        res = PaletteGroup(self.panel)
        res.grid = grid
        res.name = name
        res.title = title
        self.stack.add_titled(grid, name, title)
        return res

    def add_separator(self):
        widget = Gtk.Separator()
        self.grid.attach(widget, 0, self._row, 2, 1)
        self._row += 1

    def add_tabs(self, panel):
        box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)
        self.grid.attach(box, 0, self._row, 3, 1)
        self._row += 1
        stack = Gtk.Stack()
        switcher = Gtk.StackSwitcher()
        switcher.get_style_context().add_class("less-padding")
        box.pack_start(switcher, True, True, 0)
        box.pack_start(stack, True, True, 0)

        switcher.set_stack(stack)

        res = PaletteGroup(panel)
        res.stack = stack
        return res

    def add_control(self, name, widget=None, widget2=None):
        label = Gtk.Label(name, xalign=1.0)
        label.get_style_context().add_class('dim-label')
        self.grid.attach(label, 0, self._row, 1, 1)
        width = 2
        if widget2 is not None:
            width = 1
        if widget is not None:
            self.grid.attach(widget, 1, self._row, width, 1)
        if widget2 is not None:
            self.grid.attach(widget2, 2, self._row, width, 1)
        self._row += 1

    def add_control_dropdown(self, name, model, handler=None, handler_args=None):
        widget = Gtk.ComboBox.new_with_model(model)
        widget.set_entry_text_column(1)
        widget.set_id_column(0)
        self.add_control(name, widget)
        renderer = Gtk.CellRendererText()
        widget.pack_start(renderer, True)
        widget.add_attribute(renderer, "text", 1)
        if handler_args is None:
            handler_args = []
        if handler is not None:
            def wrapper(*args, **kwargs):
                if self.panel.model_changing:
                    return
                handler(*args, *handler_args, **kwargs)

            widget.connect('changed', wrapper)
        return widget

    def add_control_color(self, name, handler=None, handler_args=None):
        if handler_args is None:
            handler_args = []

        widget = Gtk.ColorButton()
        self.add_control(name, widget)
        if handler is not None:
            def wrapper(*args, **kwargs):
                if self.panel.model_changing:
                    return
                handler(*args, *handler_args, **kwargs)

            widget.connect('color-set', wrapper)

        return widget

    def add_control_toggle(self, name, label, handler=None, handler_args=None):
        # lbl = Gtk.Label(label=label)
        widget = Gtk.Button(label)
        # widget.add(lbl)
        widget.set_size_request(48, 48)
        widget.get_style_context().add_class('bmdbtn')
        self.add_control(name, widget)
        if handler_args is None:
            handler_args = []

        def wrapper(*args, **kwargs):
            handler(*args, *handler_args, **kwargs)

        if handler is not None:
            widget.connect('clicked', wrapper)
        return widget

    def add_control_slider(self, name, adjustment: Gtk.Adjustment, handler=None):
        widget = Gtk.Scale()
        widget.set_draw_value(False)
        widget.set_adjustment(adjustment)
        widget.set_hexpand(True)
        widget.connect('button-press-event', self.panel._on_slider_held)
        widget.connect('button-release-event', self.panel._on_slider_released)
        self.add_control(name, widget)
        if handler is not None:
            def wrapper(*args, **kwargs):
                if not self.panel.slider_held:
                    return
                handler(*args, **kwargs)

            adjustment.connect('value-changed', wrapper)

        return widget

    def add_control_slider_box(self, name, adjustment: Gtk.Adjustment, handler=None, handler_args=None,
                               display_min=None, display_max=None):
        widget = Gtk.Scale()
        widget.set_draw_value(False)
        widget.set_adjustment(adjustment)
        widget.connect('button-press-event', self.panel._on_slider_held)
        widget.connect('button-release-event', self.panel._on_slider_released)
        widget.set_hexpand(True)

        if display_min is None:
            display_min = adjustment.get_lower()
        if display_max is None:
            display_max = adjustment.get_upper()

        box = AdjustmentEntry(adjustment, display_min, display_max)
        box.set_size_request(24, 24)
        box.set_hexpand(False)
        box.set_halign(Gtk.Align.END)
        box.set_width_chars(6)

        self.add_control(name, widget, box)
        if handler_args is None:
            handler_args = []
        if handler is not None:
            def wrapper(*args, **kwargs):
                if not self.panel.slider_held:
                    return
                handler(*args, *handler_args, **kwargs)

            adjustment.connect('value-changed', wrapper)

        return widget

    def add_control_radio(self, name, options, handler=None, handler_args=None):
        box = Gtk.Box(orientation=Gtk.Orientation.VERTICAL)

        if handler_args is None:
            handler_args = []

        def wrapper(*args, **kwargs):
            if self.panel.model_changing:
                return
            handler(*args, args[0].value, *handler_args, **kwargs)

        group = None
        radios = []
        for key in options:
            rad = Gtk.RadioButton(label=options[key], group=group)
            rad.value = key
            if group is None:
                group = rad
            if handler is not None:
                rad.connect("clicked", wrapper)
            box.pack_start(rad, True, True, 0)
            radios.append(rad)
        self.add_control(name, box)
        return RadioProxy(self.panel, radios)


class PalettePanel(Gtk.Overlay):
    __gtype_name__ = 'PalettePanel'

    def __init__(self, panel_name, connection: AtemProtocol, preset_domain=None, preset_override=None,
                 preset_fields=None):
        self.panel_name = panel_name
        self.connection = connection
        self.window = None
        self.application = None

        self.model_changing = False
        self.model = EventGate(self)
        self.slider_held = False

        self._row = 0
        self._toplevel = PaletteGroup(self)

        super(Gtk.Overlay, self).__init__()

        # Build expander for panel
        self.expander = Gtk.Expander()
        self.expander.get_style_context().add_class('bmdgroup')
        frame_label = Gtk.Label(self.panel_name)
        frame_label.get_style_context().add_class("heading")
        self.expander.set_label_widget(frame_label)
        self.expander.set_expanded(True)
        self.add(self.expander)

        # Optionally create the preset button for this panel
        self.preset_domain = None
        self.preset_override = preset_override
        if preset_domain is not None:
            self.preset_fields = preset_fields
            self.preset_domain = preset_domain
            self.preset_button = Gtk.MenuButton()
            self.preset_button.set_valign(Gtk.Align.START)
            self.preset_button.set_halign(Gtk.Align.END)
            self.preset_button.set_size_request(28, 28)
            self.set_class(self.preset_button, 'flat', True)
            self.set_class(self.preset_button, 'preset', True)
            self.add_overlay(self.preset_button)

            hamburger = Gtk.Image.new_from_icon_name('open-menu-symbolic', Gtk.IconSize.BUTTON)
            self.preset_button.set_image(hamburger)
            self.preset_button.set_name(preset_domain)

            # Create `panel` action group
            self._action_group = str(uuid.uuid4())
            agp = Gio.SimpleActionGroup()
            self.insert_action_group(self._action_group, agp)

            # Create actions for the panel
            action = Gio.SimpleAction.new("savepreset", None)
            action.connect("activate", self._on_save_preset)
            agp.add_action(action)

            action = Gio.SimpleAction.new("recallpreset", GLib.VariantType.new("(ss)"))
            action.connect("activate", self._on_recall_preset)
            agp.add_action(action)

            self._preset_menu = Gio.Menu()
            self._presets = {}
            self._preset_submenu = Gio.Menu()
            self._preset_menu.append('Save preset', f"{self._action_group}.savepreset")
            self._preset_menu.append_section('Presets', self._preset_submenu)
            self._load_presets()
            self.preset_button.set_menu_model(self._preset_menu)

        # Create the expandable frame
        self.frame = Gtk.Frame()
        self.expander.add(self.frame)
        self.frame.set_margin_top(6)
        self.frame.get_style_context().add_class('view')

        # Create the grid to contain the settings
        self.grid = Gtk.Grid()
        self.grid.set_column_spacing(12)
        self.grid.set_row_spacing(12)
        self.grid.set_margin_top(12)
        self.grid.set_margin_bottom(12)
        self.grid.set_margin_start(12)
        self.grid.set_margin_end(12)
        self.frame.add(self.grid)
        self._toplevel.grid = self.grid

        self.show_all()

        # Hook up decorators
        ptype = self.__class__.__name__
        if ptype in _palette_field:
            callbacks = _palette_field[ptype]
            for selector in callbacks:
                for fun in callbacks[selector]:
                    fn_ref = getattr(self, fun)
                    self.connection.on(selector, fn_ref)

        GLib.timeout_add_seconds(1, self.initialize)

    def __repr__(self):
        return '<PalettePanel {}>'.format(self.panel_name)

    def _init_flatten(self, state):
        result = []
        if isinstance(state, dict):
            for key in state:
                result.extend(self._init_flatten(state[key]))
        else:
            result.append(state)
        return result

    def initialize(self):
        ptype = self.__class__.__name__
        if ptype in _palette_field:
            callbacks = _palette_field[ptype]
            for selector in callbacks:
                if not selector.startswith('change:'):
                    continue
                part = selector.split(":")[1:]
                if part[0] not in self.connection.mixerstate:
                    return False
                values = self._init_flatten(self.connection.mixerstate[part[0]])
                for fun in callbacks[selector]:
                    fn_ref = getattr(self, fun)
                    for v in values:
                        fn_ref(v)

        return False

    def _load_presets(self):
        xdg_config_home = os.path.expanduser(os.environ.get('XDG_CONFIG_HOME', '~/.config'))
        presetdir = os.path.join(xdg_config_home, 'openswitcher', 'presets', self.preset_domain)
        loaded = {}
        for existing in glob.glob(os.path.join(presetdir, '*.json')):
            preset_code = os.path.basename(existing).replace('.json', '')
            with open(existing, 'r') as handle:
                data = json.load(handle)
                self._presets[preset_code] = data['fields']
                loaded[data['name']] = preset_code
        ordered = list(sorted(loaded.keys()))
        for preset_name in ordered:
            preset_code = loaded[preset_name]
            mi = Gio.MenuItem.new(preset_name, f"{self._action_group}.recallpreset")
            mi.set_detailed_action(f"{self._action_group}.recallpreset(('{self.preset_domain}', '{preset_code}'))")
            self._preset_submenu.append_item(mi)

    def _on_recall_preset(self, action, parameters):
        parameters = tuple(parameters)
        presetcode = parameters[1]
        preset = self._presets[presetcode]
        cmds = []

        for field in preset:
            classname = field['_name'].title().replace('-', '') + "Field"
            if hasattr(fieldmodule, classname):
                fieldclass = getattr(fieldmodule, classname)
                cmds.extend(fieldclass.restore(field, instance_override=self.preset_override))
            else:
                self.log_aw.error(f"Unknown field in preset: {field['_name']}")
                return
        self.connection.send_commands(cmds)

    def _filter_preset_state(self, field, state, override):
        result = []
        if override is None:
            override = []
        if isinstance(state, dict):
            if len(override) > 0:
                next_override = override[1:]
            else:
                next_override = []
            for key in state:
                if len(override) > 0 and override[0] != key:
                    continue
                result.extend(self._filter_preset_state(field, state[key], next_override))
        else:
            serialized = state.serialize()
            if serialized is not None:
                serialized['_name'] = field
                result.append(serialized)
        return result

    def _on_save_preset(self, action, parameters):
        contents = []
        for field in self.preset_fields:
            ms = self.connection.mixerstate[field]
            contents.extend(self._filter_preset_state(field, ms, self.preset_override))

        dialog = PresetWindow(self.window, self.application)
        response = dialog.run()

        if response == Gtk.ResponseType.CANCEL:
            return

        preset_name = dialog.get_name()
        dialog.destroy()

        code = str(uuid.uuid4())
        self._presets[code] = contents
        mi = Gio.MenuItem.new(preset_name, f"{self._action_group}.recallpreset")
        mi.set_detailed_action(f"{self._action_group}.recallpreset(('{self.preset_domain}', '{code}'))")
        self._preset_submenu.append_item(mi)

        xdg_config_home = os.path.expanduser(os.environ.get('XDG_CONFIG_HOME', '~/.config'))
        presetdir = os.path.join(xdg_config_home, 'openswitcher', 'presets', self.preset_domain)
        os.makedirs(presetdir, exist_ok=True)
        presetfile = os.path.join(presetdir, f'{code}.json')
        with open(presetfile, 'w') as handle:
            json.dump({
                'name': preset_name,
                'fields': contents,
            }, handle)

    def set_class(self, widget, classname, state):
        if state:
            widget.get_style_context().add_class(classname)
        else:
            widget.get_style_context().remove_class(classname)

    def add_control(self, name, widget1=None, widget2=None):
        return self._toplevel.add_control(name, widget1, widget2)

    def add_tabs(self):
        return self._toplevel.add_tabs(self)

    def add_separator(self):
        return self._toplevel.add_separator()

    def add_control_dropdown(self, name, model, handler=None):
        return self._toplevel.add_control_dropdown(name, model, handler=handler)

    def add_control_toggle(self, name, label, handler=None, handler_args=None):
        return self._toplevel.add_control_toggle(name, label, handler=handler, handler_args=handler_args)

    def add_control_color(self, name, handler=None, handler_args=None):
        return self._toplevel.add_control_color(name, handler=handler, handler_args=handler_args)

    def add_control_slider(self, name, adjustment: Gtk.Adjustment, handler=None):
        return self._toplevel.add_control_slider(name, adjustment, handler=handler)

    def add_control_slider_box(self, name, adjustment: Gtk.Adjustment, handler=None, handler_args=None,
                               display_min=None, display_max=None):
        return self._toplevel.add_control_slider_box(name, adjustment, handler=handler, handler_args=handler_args,
                                                     display_min=display_min, display_max=display_max)

    def run(self, cmd):
        """ Dispatch CMD to active connection """
        self.connection.send_commands([cmd])

    def _on_slider_held(self, *args):
        self.slider_held = True

    def _on_slider_released(self, *args):
        self.slider_held = False

    @classmethod
    def event(cls, name):
        """ Register the class name and function name in a global so on initialisation of the class
        the function can be hooked up to the active ATEM connection """

        def wrapper(func):
            global _palette_field
            panel, fn_name = func.__qualname__.split('.', maxsplit=1)
            if panel not in _palette_field:
                _palette_field[panel] = {}
            if name not in _palette_field[panel]:
                _palette_field[panel][name] = []

            _palette_field[panel][name].append(fn_name)
            return func

        return wrapper



# Copyright 2021 - 2022, Martijn Braam and the OpenAtem contributors
# SPDX-License-Identifier: GPL-3.0-only
import math

import gi

from gtk_switcher.template_i18n import TemplateLocale
from gtk_switcher.videohubconnection import VideoHubConnection

gi.require_version('Gtk', '3.0')

from gi.repository import Gtk, GObject


@TemplateLocale(resource_path='/nl/brixit/switcher/ui/videohub-bus.glade')
class VideoHubBus(Gtk.Grid):
    __gtype_name__ = 'VideoHubBus'
    aux_name = Gtk.Template.Child()
    bus = Gtk.Template.Child()
    focus_dummy = Gtk.Template.Child()

    def __init__(self, provider, connection, output):
        super(Gtk.Grid, self).__init__()
        self.provider = provider
        if not isinstance(connection, VideoHubConnection):
            raise ValueError()

        self.init_template()
        self.connection = connection
        self.output = int(output)
        self.aux_name.set_label(connection.outputs[self.output]['label'])
        self.create_buttons()
        self.connection._route_change.append(self.on_route_change)

    def set_class(self, widget, classname, state):
        if state:
            widget.get_style_context().add_class(classname)
        else:
            widget.get_style_context().remove_class(classname)

    def apply_css(self, widget, provider):
        Gtk.StyleContext.add_provider(widget.get_style_context(),
                                      provider,
                                      Gtk.STYLE_PROVIDER_PRIORITY_APPLICATION)

        if isinstance(widget, Gtk.Container):
            widget.forall(self.apply_css, provider)

    def create_buttons(self):
        shortnames = {
            'Black': 'BLK',
            'Color Bars': 'BARS',
            'Color 1': 'COL1',
            'Color 2': 'COL2',
            'Media Player 1': 'MP1',
            'Media Player 1 Key': 'MP1K',
            'Media Player 2': 'MP2',
            'Media Player 2 Key': 'MP2K',
            'Multi View': 'MVW',
            'Program': 'PGM',
            'Preview': 'PVW',
            'Computer Direct': 'DIR',
        }

        rows = len(self.connection.inputs) // 10
        cols = math.ceil(len(self.connection.inputs) / rows)

        left = 0
        top = 0
        for index in self.connection.inputs:
            name = self.connection.inputs[index]['label']
            if ':' in name:
                short_name, name = name.split(':', maxsplit=1)
                short_name = short_name.strip()
            elif name in shortnames:
                short_name = shortnames[name]
            else:
                short_name = str(index)

            label = Gtk.Label(label=short_name)
            btn = Gtk.Button()
            btn.add(label)
            btn.index = index
            btn.set_size_request(48, 48)
            btn.get_style_context().add_class('bmdbtn')
            if index == self.connection.outputs[self.output]['source']:
                btn.get_style_context().add_class('program')
            btn.connect('clicked', self.do_source_change)
            self.bus.attach(btn, left, top, 1, 1)
            left += 1
            if left > cols - 1:
                left = 0
                top += 1
        self.apply_css(self.bus, self.provider)
        self.bus.show_all()

    def on_route_change(self, hub_id, index, source):
        if int(index) == self.output:
            for btn in self.bus:
                if btn.index == int(source):
                    btn.get_style_context().add_class('program')
                else:
                    btn.get_style_context().remove_class('program')

    def do_source_change(self, widget):
        self.focus_dummy.grab_focus()
        self.connection.change_route(self.output, widget.index)



# Copyright 2021 - 2022, Martijn Braam and the OpenAtem contributors
# SPDX-License-Identifier: GPL-3.0-only
import json

from gi.repository import Gtk, GObject, Gdk

from gtk_switcher.template_i18n import TemplateLocale

import pyatem.videohub as videohubprotocol


@TemplateLocale(resource_path='/nl/brixit/switcher/ui/videohub.glade')
class VideoHub(Gtk.Grid):
    __gtype_name__ = 'VideoHub'

    ip_entry = Gtk.Template.Child()
    connectbtn = Gtk.Template.Child()
    status = Gtk.Template.Child()
    outputs = Gtk.Template.Child()

    def __init__(self):
        super(Gtk.Grid, self).__init__()
        self.init_template()

        self.probed = False
        self.input_model = None
        self.old_ip = None

        self.config = {
            'ip': None,
            'outputs': {}
        }

    def set_class(self, widget, classname, state):
        if state:
            widget.get_style_context().add_class(classname)
        else:
            widget.get_style_context().remove_class(classname)

    def set_input_model(self, model):
        self.input_model = model

    @GObject.Signal(name="config-changed", flags=GObject.SignalFlags.RUN_LAST, return_type=bool,
                    arg_types=(str,),
                    accumulator=GObject.signal_accumulator_true_handled)
    def settings_changed(self, *args):
        pass

    @GObject.Signal(name="ip-changed", flags=GObject.SignalFlags.RUN_LAST, return_type=bool,
                    arg_types=(str, str),
                    accumulator=GObject.signal_accumulator_true_handled)
    def ip_changed(self, *args):
        pass

    @GObject.Signal(name="deleted", flags=GObject.SignalFlags.RUN_LAST, return_type=bool,
                    arg_types=(str,),
                    accumulator=GObject.signal_accumulator_true_handled)
    def deleted(self, *args):
        pass

    def load(self, config):
        self.config = config
        self.ip_entry.set_text(config['ip'])
        self.old_ip = config['ip']
        self.on_connect_clicked(None, None)

    def save(self):
        self.emit('config-changed', json.dumps(self.config))

    @Gtk.Template.Callback()
    def on_delete_clicked(self, widget, *args):
        self.emit('deleted', self.old_ip)

    @Gtk.Template.Callback()
    def on_connect_clicked(self, widget, *args):
        ip = self.ip_entry.get_text()
        if ip.strip() == "":
            return

        if ip != self.old_ip:
            self.emit('ip-changed', self.old_ip, ip)
            self.old_ip = ip

        self.status.set_text(_("Connecting..."))

        self.probed = False
        probe = videohubprotocol.VideoHub(ip)
        probe.on('connect', self.on_connect)
        try:
            probe.connect()
        except OSError as e:
            self.status.set_text(str(e))
            return
        while not self.probed:
            probe.loop()

        self.status.set_text(_("Connected") + f" ({probe.model_display})")
        self.config['ip'] = ip
        for output in self.outputs:
            self.outputs.remove(output)

        for index in probe.output_label:
            output = probe.output_label[index]
            row = Gtk.Box(spacing=8)

            # Output label
            label = Gtk.Label(output)
            label.set_margin_start(8)
            row.add(label)

            # Show the videohub output as bus, checkbox
            checkbox = Gtk.CheckButton(_("Show as bus"))
            if str(index) in self.config['outputs']:
                checkbox.set_active(self.config['outputs'][str(index)]['bus'])
            checkbox.output = index
            checkbox.set_margin_end(8)
            checkbox.connect('toggled', self.on_bus_toggled)
            row.pack_end(checkbox, False, False, 0)

            # Source label checkbox
            checkbox = Gtk.CheckButton(_("Rename on change"))
            if str(index) in self.config['outputs']:
                checkbox.set_active(self.config['outputs'][str(index)]['rename'])
            checkbox.output = index
            checkbox.connect('toggled', self.on_rename_toggled)
            row.pack_end(checkbox, False, False, 0)

            # Atem input link
            input_select = Gtk.ComboBox.new_with_model(self.input_model)
            input_select.set_entry_text_column(1)
            input_select.set_id_column(0)
            if str(index) in self.config['outputs']:
                if self.config['outputs'][str(index)]['source'] is not None:
                    input_select.set_active_id(str(self.config['outputs'][str(index)]['source']))
            input_select.output = index
            input_select.connect('changed', self.on_source_change)
            renderer = Gtk.CellRendererText()
            input_select.pack_start(renderer, True)
            input_select.add_attribute(renderer, "text", 1)

            row.pack_end(input_select, False, False, 0)

            self.outputs.add(row)

            if str(index) not in self.config['outputs']:
                self.config['outputs'][str(index)] = {
                    'bus': False,
                    'rename': False,
                    'source': None,
                }

        self.outputs.show_all()
        self.save()

    def on_connect(self, *args):
        self.probed = True

    def on_source_change(self, widget, *args):
        output = widget.output
        index = widget.get_active_id()
        if index == "":
            index = None
        else:
            index = int(index)

        self.config["outputs"][str(output)]["source"] = index
        self.save()

    def on_bus_toggled(self, widget, *args):
        output = widget.output
        self.config["outputs"][str(output)]["bus"] = widget.get_active()
        self.save()

    def on_rename_toggled(self, widget, *args):
        output = widget.output
        self.config["outputs"][str(output)]["rename"] = widget.get_active()
        self.save()



import argparse
import os
import sys
import logging
import tqdm
from pathlib import Path

from PIL import Image

from pyatem.command import CutCommand, TimeRequestCommand
import pyatem.mediaconvert
import pyatem.media
from pyatem.cameracontrol import CameraControlData
from pyatem.protocol import AtemProtocol
import pyatem.field as fieldmodule

connection = None
slot_index = None
image_path = None
args = None
pbar = None
logging.basicConfig(level=logging.INFO)


def prepare_image(path, resolution):
    im = Image.open(path)
    frame = Image.new('RGBA', resolution)
    im.thumbnail(resolution, Image.Resampling.LANCZOS)
    frame.paste(im)
    pixels = frame.getdata()

    flat = [item for sublist in pixels for item in sublist]
    return bytes(flat)


def save_image(path, resolution, data):
    im = Image.frombytes('RGBA', resolution, data)
    im.save(path)


def changed(key, contents):
    if key == 'time':
        return
    if not args.debug:
        return
    if isinstance(contents, fieldmodule.FieldBase):
        print(contents)
    else:
        print(key)


def connected():
    if not isinstance(connection, AtemProtocol):
        raise ValueError()

    product = connection.mixerstate['product-name']
    slots = connection.mixerstate['mediaplayer-slots']
    mode = connection.mixerstate['video-mode']

    logging.info(f'Connected to {product.name} at {mode.get_label()}')

    if slot_index < 0 or slot_index > slots.stills:
        logging.fatal(f'Slot index out of range, This hardware supports slot 1-{slots.stills}')

    if args.action == "upload":
        frame = prepare_image(image_path, mode.get_resolution())
        connection.send_commands([TimeRequestCommand()])
        logging.basicConfig(level=logging.DEBUG)
        frame_atem = pyatem.media.rgb_to_atem(frame, *mode.get_resolution())
        if args.name:
            name = args.name
        else:
            name = Path(args.file).stem
        connection.upload(0, slot_index, frame_atem, name=name, compress=True)
    elif args.action == "download":
        connection.download(0, slot_index)


def uploaded(store, slot):
    logging.info("Upload completed")
    exit(0)


def downloaded(store, slot, data):
    logging.info(f'Download complete, received {len(data)} bytes')
    if args.raw:
        logging.info(f'Saving raw data to {args.file}')
        with open(args.file, 'wb') as handle:
            handle.write(data)
    else:
        mode = connection.mixerstate['video-mode']
        image = pyatem.media.atem_to_image(data, *mode.get_resolution())
        save_image(args.file, mode.get_resolution(), image)
    exit(0)


def progress(store, slot, factor):
    print(factor * 100)


def upload_progress(store, slot, percent, done, size):
    global pbar
    if pbar is None:
        pbar = tqdm.tqdm(total=size, unit='B', unit_scale=True)
        pbar.last_done = 0
    block = done - pbar.last_done
    pbar.update(block)
    pbar.last_done = done
    if done == size:
        pbar.close()


def main():
    global connection, slot_index, image_path, args
    parser = argparse.ArgumentParser()
    parser.add_argument('--debug', help="More debug output", action="store_true")
    subparsers = parser.add_subparsers(dest="action")
    subparsers.required = True
    upload_parser = subparsers.add_parser('upload')
    upload_parser.add_argument('ip', help='ATEM IP address')
    upload_parser.add_argument('index', help='Media store slot number', type=int)
    upload_parser.add_argument('file', help='Still to upload')
    upload_parser.add_argument('--name', help='Still name')
    download_parser = subparsers.add_parser('download')
    download_parser.add_argument('ip', help='ATEM IP address')
    download_parser.add_argument('index', help='Media store slot number', type=int)
    download_parser.add_argument('file', help='Local filename for the still')
    download_parser.add_argument('--raw', help='Don\'t decode', action='store_true')

    args = parser.parse_args()

    if args.action == "upload" and not os.path.isfile(args.file):
        sys.stderr.write('File not found\n')
        exit(1)

    slot_index = args.index - 1
    image_path = args.file

    logging.info(f'Connecting to ATEM at {args.ip}...')
    if args.ip == 'usb':
        connection = AtemProtocol(usb=True)
    else:
        connection = AtemProtocol(args.ip)
    connection.on('connected', connected)
    connection.on('upload-done', uploaded)
    connection.on('download-done', downloaded)
    connection.on('transfer-progress', progress)
    connection.on('upload-progress', upload_progress)
    connection.on('change', changed)

    connection.connect()
    while True:
        connection.loop()


if __name__ == '__main__':
    main()



# Copyright 2021 - 2022, Martijn Braam and the OpenAtem contributors
# SPDX-License-Identifier: GPL-3.0-only
import math

import gi

gi.require_version('Gtk', '3.0')
from gi.repository import Gtk, GLib, GObject, Gio, Gdk


class Dial(Gtk.Range):
    __gtype_name__ = 'Dial'

    def __init__(self):
        super(Gtk.Range, self).__init__()
        self.set_size_request(48, 48)
        self.editing = False
        self.start_x = None
        self.position = None
        self.connect('button-press-event', self.on_mouse_down)
        self.connect('button-release-event', self.on_mouse_up)
        self.connect('motion-notify-event', self.on_mouse_move)

    def _polar(self, dist, angle):
        x = dist * math.cos(math.radians(angle))
        y = dist * math.sin(math.radians(angle))
        return x, y

    def do_draw(self, cr):
        allocation = self.get_allocation()
        context = self.get_style_context()

        context.save()
        context.add_class('dial-frame')
        Gtk.render_background(context, cr, 0, 0, allocation.width, allocation.height)
        Gtk.render_frame(context, cr, 0, 0, allocation.width, allocation.height)
        context.restore()
        context.save()
        context.add_class('dial')
        if self.editing:
            context.add_class('active')

        padding = 16
        top = padding / 2
        left = padding / 2
        diameter = min(allocation.width - padding, allocation.height - padding)
        if allocation.width - padding > diameter:
            left = (allocation.width - diameter) // 2

        cx = allocation.width / 2
        cy = allocation.height / 2

        Gtk.render_background(context, cr, left, top, diameter, diameter)
        Gtk.render_frame(context, cr, left, top, diameter, diameter)
        context.restore()

        context.save()
        context.add_class('dial-mark')

        value = self.get_value()
        range_min = self.get_adjustment().get_lower()
        range_max = self.get_adjustment().get_upper()
        if range_max - range_min == 0:
            value = 0
        else:
            value = ((value - range_min) / (range_max - range_min))
        inner = self._polar(diameter * 0.2, (value * 270) + 45 + 90)
        outer = self._polar(diameter * 0.4, (value * 270) + 45 + 90)
        cr.set_line_width(2)
        cr.move_to(cx + inner[0] + 0.5, cy + inner[1] + 0.5)
        cr.line_to(cx + outer[0] + 0.5, cy + outer[1] + 0.5)
        cr.set_source_rgba(255, 255, 255, 1)
        cr.stroke()
        context.restore()

    def on_mouse_down(self, widget, event):
        self.editing = True
        self.start_x = event.x
        adj = self.get_adjustment()
        value = adj.get_value()
        self.position = self._remap(value, adj.get_lower(), adj.get_upper(), -100, 100)

    def on_mouse_up(self, widget, *args):
        self.editing = False
        self.invalidate()

    def on_mouse_move(self, widget, event):
        if self.editing:
            new_x = event.x

            change = new_x - self.start_x
            change += self.position
            new_val = self._remap(change, -100, 100, self.get_adjustment().get_lower(),
                                  self.get_adjustment().get_upper())
            self.get_adjustment().set_value(new_val)

    def _remap(self, value, old_min, old_max, new_min, new_max):
        return ((value - old_min) / (old_max - old_min)) * (new_max - new_min) + new_min

    def invalidate(self):
        # TODO: Make functional
        return
        rect = self.get_allocation()
        parent = self.get_parent()
        if parent:
            offset = parent.get_allocation()
            rect.x -= offset.x
            rect.y -= offset.y

        self.window.invalidate_rect(rect, False)



import hmac
import struct
import threading
import logging
import socketserver
from functools import partial

from pyatem.command import TransferCompleteCommand
from pyatem.field import InitCompleteField
from pyatem.transfer import TransferTask


class TCPHandler(socketserver.BaseRequestHandler):
    def __init__(self, config, threadpool, *args, **kwargs):
        self.config = config
        self.threadpool = threadpool
        self.device = None
        self.callback_id = None
        self.callback_uploaded = None
        self.transfer_buffer = {}
        super().__init__(*args, **kwargs)

    def setup(self):
        if not hasattr(self.server, 'numclients'):
            self.server.numclients = 0
        self.server.numclients += 1

    def decode_packet(self, data):
        offset = 0
        if len(data) < 8:
            raise ValueError("Packet too short")
        while offset < len(data):
            datalen, cmd = struct.unpack_from('!H2x 4s', data, offset)
            if datalen > 8 + 32:
                raise ValueError("Received large packet")
            raw = data[offset + 8:offset + datalen]
            yield (cmd, raw)
            offset += datalen

    def list_to_packets(self, data):
        result = b''
        for key, value in data:
            result += struct.pack('!H2x 4s', len(value) + 8, key)
            result += value
        return result

    def _flatten(self, idict):
        result = []
        for key in idict:
            if isinstance(idict[key], dict):
                result.extend(self._flatten(idict[key]))
            elif isinstance(idict[key], list):
                result.extend(idict[key])
            else:
                result.append(idict[key])
        return result

    def send_packets(self, data):
        data = self.list_to_packets(data)
        self.send_raw(data)

    def send_raw(self, data):
        header = struct.pack('!H', len(data))
        self.request.sendall(header + data)

    def send_fields(self, fields):
        data = b''
        for field in fields:
            data += field.make_packet()
        header = struct.pack('!H', len(data))
        self.request.sendall(header + data)

    def send_initial_sync(self):
        mixerstate = self.threadpool['hardware'][self.device].switcher.mixerstate
        state = self._flatten(mixerstate)

        buffer = []
        size = 0
        for field in state:
            if isinstance(field, bytes):
                continue
            fsize = len(field.raw) + 8
            if size + fsize > 1000:
                self.send_fields(buffer)
                buffer = []
                size = 0
            buffer.append(field)
            size += fsize
        self.send_fields(buffer)

        buffer = [InitCompleteField(b'\0\0\0\0')]
        self.send_fields(buffer)

    def receive(self):
        try:
            header = self.request.recv(2)
        except ConnectionResetError:
            raise ValueError("Client disconnected")
        if len(header) == 0:
            raise ValueError("Client disconnected")
        datalength, = struct.unpack('!H', header)
        data_left = datalength
        data = b''
        while data_left > 0:
            block = self.request.recv(data_left)
            if len(block) == 0:
                logging.warning("Connection closed")
                return
            data_left -= len(block)
            data += block

        return data

    def handle(self):
        # Rename thread
        t = threading.currentThread()
        t.setName('tcp.{}:{}'.format(self.client_address[0], self.client_address[1]))

        try:
            # Handshake magic packet
            handshake = self.receive()
            packets = list(self.decode_packet(handshake))
            if len(packets) > 1:
                logging.warning('Too many packets in handshake, rejecting')
                return

            if packets[0][0] != b'*SW*':
                logging.warning('Invalid magic on new connection, rejecting')
                return

            # Optionally run the auth
            if self.config['auth']:
                self.send_packets([(b'AUTH', b'')])
                raw = self.receive()
                packets = list(self.decode_packet(raw))
                fields = {}
                for packet in packets:
                    fields[packet[0]] = packet[1]
                username = fields[b'*USR'].decode()
                password = fields[b'*PWD'].decode()
                user_ok = hmac.compare_digest(self.config['username'], username)
                pass_ok = hmac.compare_digest(self.config['password'], password)
                if not user_ok or not pass_ok:
                    logging.warning("Invalid login information supplied, rejecting")
                    return

            # Send device list to client
            logging.info('Client connected')
            hardware = []
            for key in self.config['hardware'].split(','):
                label = self.threadpool['hardware'][key].config['label']
                hardware.append((b'*HW*', struct.pack('>20s20s', key.encode(), label.encode())))
            self.send_packets(hardware)

            # Device selection
            raw = self.receive()
            packets = list(self.decode_packet(raw))
            if packets[0][0] != b'*DEV':
                logging.error('Expected *DEV response, rejecting')
                return
            self.device = packets[0][1].decode()
            logging.info('selected device ' + str(self.device))

            # Initial sync
            self.send_initial_sync()

            # Register events
            self.callback_id = self.threadpool['hardware'][self.device].switcher.on('change', self.proxy_change)
            self.callback_upload = self.threadpool['hardware'][self.device].switcher.on('upload-done',
                                                                                        self.proxy_uploaded)

            # Proxying
            while True:
                packet = self.receive()
                pid = packet[4:8]
                if pid.startswith(b'*'):
                    cmd = pid[1:4].decode()
                    handler_name = f'handle_{cmd.lower()}'
                    if hasattr(self, handler_name):
                        handler = getattr(self, handler_name)
                        handler(packet)
                    else:
                        logging.error(f"Unknown command: {cmd}")
                else:
                    self.threadpool['hardware'][self.device].switcher.send_raw(packet)


        except ValueError as e:
            logging.error("Protocol error: " + str(e))
            return

    def handle_xfr(self, packet):
        task = TransferTask.from_tcp(packet)
        id = (task.store, task.slot, task.upload)
        if id not in self.transfer_buffer:
            self.transfer_buffer[id] = task
        else:
            self.transfer_buffer[id].data += task.data

        received = len(self.transfer_buffer[id].data)
        if received == task.send_length:
            logging.info(f'Uploading to store {task.store} slot {task.slot}')
            hw = self.threadpool['hardware'][self.device].switcher
            if task.upload:
                hw.upload(task.store, task.slot, b'', task=self.transfer_buffer[id])

    def finish(self):
        if self.callback_id is not None:
            self.threadpool['hardware'][self.device].switcher.off('change', self.callback_id)
        if self.callback_upload is not None:
            self.threadpool['hardware'][self.device].switcher.off('upload-done', self.callback_upload)
        self.server.numclients -= 1

    def proxy_change(self, key, val):
        if isinstance(val, bytes):
            pass
            # Don't send packets we can't decode yet
            # self.send_raw(val)
        else:
            self.send_raw(val.make_packet())

    def proxy_uploaded(self, store, slot):
        del self.transfer_buffer[(store, slot, True)]
        self.send_raw(TransferCompleteCommand(store, slot, True).get_command())


class TcpFrontendThread(threading.Thread):
    def __init__(self, config, threadlist):
        threading.Thread.__init__(self)
        self.name = 'tcp.' + str(config['bind'])
        self.config = config
        self.threadlist = threadlist
        self.stop = False
        self.server = None

    def run(self):
        logging.info('TCP frontend run')
        host, port = self.config['bind'].split(':')
        address = (host, int(port))
        logging.info(f'binding to {address}')

        socketserver.TCPServer.allow_reuse_address = True
        handler = partial(TCPHandler, self.config, self.threadlist)
        self.server = socketserver.ThreadingTCPServer(address, handler)
        self.server.numclients = 0
        self.server.serve_forever()

    def get_status(self):
        return 'running, {} clients'.format(self.server.numclients)
